import inspect
import warnings
from datetime import datetime
from typing import TYPE_CHECKING, List, Optional, Union
from onetick.py.backports import Literal

from onetick import py as otp
from onetick.py import types as ott
from onetick.py.core._internal._state_objects import _TickSequence
from onetick.py.core._source._symbol_param import _SymbolParamSource
from onetick.py.core.column_operations._methods.op_types import are_strings
from onetick.py.core.column_operations.base import _Operation
from onetick.py.core.eval_query import prepare_params
from onetick.py.otq import otq
from onetick.py.compatibility import (
    is_supported_point_in_time,
    is_join_with_snapshot_snapshot_fields_parameter_supported,
)

if TYPE_CHECKING:
    from onetick.py.core.source import Source


def _process_keep_time_param(self: 'Source', keep_time, sub_source):
    if keep_time == "TIMESTAMP":
        raise ValueError("TIMESTAMP is reserved OneTick name, please, specify another one.")
    if keep_time in self.columns():
        raise ValueError(f"{keep_time} column is already presented.")
    sub_source = sub_source.copy()
    if keep_time:
        sub_source[keep_time] = sub_source["Time"]
    return sub_source


def _process_start_or_end_of_jwq(join_params, time, param_name):
    if time is not None:
        if isinstance(time, (datetime, otp.dt)):
            join_params[f"{param_name}"] = ott.datetime2expr(time)
        elif isinstance(time, _Operation):
            join_params[f"{param_name}"] = str(time)
        else:
            raise ValueError(f"{param_name} should be datetime.datetime instance or OneTick expression")


def _columns_to_params_for_joins(columns, query_params=False):
    """
    Converts a dictionary of columns into a parameters string.
    This is mainly used for join_with_query and join_with_collection.

    query_params control whether resulting string should be considered query params or symbol params
    (as it impacts some of the conversion rules)
    """
    params_list = []

    def get_msecs_expression(value):
        return f"tostring(GET_MSECS({str(value)}))"

    for key, value in columns.items():
        dtype = ott.get_object_type(value)
        convert_rule = "'" + key + "=' + "

        if key == '_PARAM_SYMBOL_TIME' and not query_params:
            # this symbol parameter has to be formatted differently because Onetick treats this parameter
            # in a special way
            if dtype is otp.nsectime:
                convert_rule += f'NSECTIME_FORMAT("%Y%m%d%H%M%S.%J",{ott.value2str(value)},_TIMEZONE)'
            elif dtype is str:
                if are_strings(getattr(value, "dtype", None)):
                    convert_rule += str(value)
                else:
                    convert_rule += '"' + value + '"'
            else:
                raise ValueError('Parameter symbol_time has to be a datetime value!')

        elif dtype is str:
            if are_strings(getattr(value, "dtype", None)):
                convert_rule += str(value)
            else:
                convert_rule += '"' + value + '"'
        elif dtype is otp.msectime:
            convert_rule += get_msecs_expression(value)
        elif dtype is otp.nsectime:
            if key == '_SYMBOL_TIME' and query_params:
                # hack to support passing _SYMBOL_TIME to called query as a parameter
                warnings.warn(
                    'Query parameter _SYMBOL_TIME passed to join_with_query! '
                    'This is deprecated. Please use a dedicated `symbol_time` parameter of the '
                    'join_with_query function',
                    FutureWarning,
                    stacklevel=3,
                )
                convert_rule += get_msecs_expression(value)
            elif query_params:
                # this can be used for query params but cannot be used for symbol params
                # overall it's better
                convert_rule += "'NSECTIME('+tostring(NSECTIME_TO_LONG(" + str(value) + "))+')'"
            else:
                # this matches the common way onetick converts nanoseconds to symbol parameters
                convert_rule += (
                    get_msecs_expression(value) + "+'.'+SUBSTR(NSECTIME_FORMAT('%J'," + str(value) + ",_TIMEZONE),3,6)"
                )
        else:
            convert_rule += "tostring(" + str(value) + ")"
        params_list.append(convert_rule)
    return "+','+".join(params_list)


def _check_and_convert_symbol(symbol):
    """
    Convert the value of 'symbol' function parameter to symbol name
    OneTick string representation and dictionary of symbol parameters.
    """
    # "symbol" parameter can contain a symbol name (string, field, operation etc),
    # a symbol parameter list (dict, Source, _SymbolParamSource),
    # or both together as a tuple

    symbol_name = None
    symbol_param = {}

    # if "symbol" is tuple, we unpack it
    if isinstance(symbol, tuple) and len(symbol) == 2:
        symbol, symbol_param = symbol

    if isinstance(symbol, _Operation):  # TODO: PY-35
        symbol_name = f"tostring({str(symbol)})"
    elif isinstance(symbol, str):
        symbol_name = f"'{symbol}'"
    elif type(symbol) in {int, float}:  # constant
        symbol_name = f"tostring({symbol})"
    elif symbol is None:
        # this is necessary to distinguish None (which is valid value for symbol) from invalid values
        symbol_name = None
    else:
        if not symbol_param:
            symbol_param = symbol

    return symbol_name, symbol_param


def _convert_symbol_param_and_columns(symbol_param):
    """
    We need to create two objects from a symbol param (a dict, a Source or a _SymbolParamSource):

    1. Dictionary of columns to generate list of symbol parameters for the JOIN_WITH_QUERY EP
    2. _SymbolParamSource object to pass to the source function if necessary
    """

    if isinstance(symbol_param, dict):
        converted_symbol_param_columns = symbol_param
        converted_symbol_param = _SymbolParamSource(
            **{key: ott.get_object_type(column) for key, column in symbol_param.items()}
        )
    elif isinstance(symbol_param, otp.Source):
        converted_symbol_param_columns = {
            field_name: symbol_param[field_name] for field_name in symbol_param.columns(skip_meta_fields=True).keys()
        }
        converted_symbol_param = symbol_param.to_symbol_param()
    elif isinstance(symbol_param, _SymbolParamSource):
        converted_symbol_param_columns = {
            field_name: symbol_param[field_name] for field_name in symbol_param.schema.keys()
        }
        converted_symbol_param = symbol_param
    else:
        return None, None

    # we want to pass all the fields to the joined query as symbol parameters,
    # except for some special fields that would override explicitly set parameters
    ignore_symbol_fields = [
        '_PARAM_START_TIME_NANOS',
        '_PARAM_END_TIME_NANOS',
    ]
    filtered_converted_symbol_param_columns = {}
    for field_name, field_value in converted_symbol_param_columns.items():
        if field_name in ignore_symbol_fields:
            warnings.warn(
                f'Special symbol parameter "{field_name}" was passed to the joined query! '
                'This parameter would be ignored. Please, use parameters of the `join_with_query` '
                'function itself to set it.',
                FutureWarning,
                stacklevel=2,
            )
        else:
            filtered_converted_symbol_param_columns[field_name] = field_value
    filtered_converted_symbol_param = _SymbolParamSource(
        **{
            field_name: field_value
            for field_name, field_value in converted_symbol_param.schema.items()
            if field_name not in ignore_symbol_fields
        }
    )
    return filtered_converted_symbol_param_columns, filtered_converted_symbol_param


def _fill_time_param_for_jwq(join_params, start_time, end_time, timezone):
    _process_start_or_end_of_jwq(join_params, start_time, "start_timestamp")
    _process_start_or_end_of_jwq(join_params, end_time, "end_timestamp")
    if timezone:
        join_params["timezone"] = f"'{timezone}'"
    else:
        join_params["timezone"] = "_TIMEZONE"  # this may break something, need to test


def _fill_aux_params_for_joins(
    join_params, caching, end_time, prefix, start_time, symbol_name, timezone, for_join_with_collection=False
):
    if symbol_name and not for_join_with_collection:
        join_params["symbol_name"] = symbol_name
    if prefix is not None:
        join_params["prefix_for_output_ticks"] = str(prefix)
    if caching:
        if for_join_with_collection:
            supported = ("per_symbol",)
        else:
            supported = ("cross_symbol", "per_symbol")
        if caching in supported:
            join_params["caching_scope"] = caching
        else:
            raise ValueError(f"Unknown value for caching param, please use None or any of {supported}.")
    _fill_time_param_for_jwq(join_params, start_time, end_time, timezone)
    if for_join_with_collection:
        del join_params['timezone']


def _get_default_fields_for_outer_join_str(default_fields_for_outer_join, how, sub_source_schema):
    """
    Default fields for outer join definition.
    Used by join_with_query() and join_with_collection()
    """
    default_fields_for_outer_join_str = ''
    if default_fields_for_outer_join:
        if how != 'outer':
            raise ValueError('The `default_fields_for_outer_join` parameter can be used only for outer join')
        for field, expr in default_fields_for_outer_join.items():
            if field not in sub_source_schema.keys():
                raise KeyError(
                    f'Field {field} is specified in `default_fields_for_outer_join` parameter, '
                    'but is not present in the joined source schema!'
                )
            if default_fields_for_outer_join_str != '':
                default_fields_for_outer_join_str += ','
            default_fields_for_outer_join_str += f'{field}={ott.value2str(expr)}'
    return default_fields_for_outer_join_str


def _get_columns_with_prefix(self: 'Source', sub_source, prefix) -> dict:
    sub_source_columns = sub_source.schema
    if prefix is None:
        prefix = ""
    if not isinstance(prefix, str):
        raise ValueError("Only string constants are supported for now.")
    new_columns = {prefix + name: dtype for name, dtype in sub_source_columns.items()}
    same_names = set(new_columns) & set(self.schema)
    if same_names:
        raise ValueError(f"After applying prefix some columns aren't unique: {', '.join(same_names)}.")
    return new_columns


def join_with_collection(
    self: 'Source',
    collection_name,
    query_func=None,
    how="outer",
    params=None,
    start=None,
    end=None,
    prefix=None,
    caching=None,
    keep_time=None,
    default_fields_for_outer_join=None,
) -> 'Source':
    """
    For each tick uses ``query_func`` to join ticks from ``collection_name`` tick collection
    (tick set, unordered tick set, tick list, or tick deque).

    Parameters
    ----------
    collection_name: str
        Name of the collection state variable from which to join ticks. Collections are the following types:
        :py:class:`TickSet <onetick.py.core._internal._state_objects.TickSet>`,
        :py:class:`TickSetUnordered <onetick.py.core._internal._state_objects.TickSetUnordered>`,
        :py:class:`TickList <onetick.py.core._internal._state_objects.TickList>` and
        :py:class:`TickDeque <onetick.py.core._internal._state_objects.TickDeque>`.

    query_func: callable
        Callable ``query_func`` should return :class:`Source`. If passed, this query will be used on ticks
        from collection before joining them.
        In this case, ``query_func`` object will be evaluated by OneTick (not python)
        for every input tick. Note that python code will be executed only once,
        so all python's conditional expressions will be evaluated only once too.

        Callable should have ``source`` parameter. When callable is called, this parameter
        will have value of a :class:`Source` object representing ticks loaded directly from the collection.
        Any operation applied to this source will be applied to ticks from the collection
        before joining them.

        Also, callable should have the parameters with names
        from ``params`` if they are specified in this method.

        If ``query_func`` is not passed, then all ticks from the collection will be joined.
    how: 'inner', 'outer'
        Type of join. If **inner**, then output tick is propagated
        only if some ticks from the collection were joined to the input tick.
    params: dict
        Mapping of the parameters' names and their values for the ``query_func``.
        :py:class:`Columns <onetick.py.Column>` can be used as a value.
    start: :py:class:`otp.datetime <onetick.py.datetime>`, :py:class:`otp.Operation <onetick.py.Operation>`
        Start time to select ticks from collection.
        If specified, only ticks in collection that have higher or equal timestamp will be processed.
        If not passed, then there will be no lower time bound for the collection ticks.
        This means that even ticks with TIMESTAMP lower than _START_TIME of the main query will be joined.
    end: :py:class:`otp.datetime <onetick.py.datetime>`, :py:class:`otp.Operation <onetick.py.Operation>`
        End time to select ticks from collection.
        If specified, only ticks in collection that have lower timestamp will be processed.
        If not passed, then there will be no upper time bound for the collection ticks.
        This means that even ticks with TIMESTAMP higher than _END_TIME of the main query will be joined.
    prefix : str
        Prefix for the names of joined tick fields.
    caching : str
        If `None` caching is disabled. You can specify caching by using values:

            * 'per_symbol': cache is different for each symbol.
    keep_time : str
        Name for the joined timestamp column. `None` means no timestamp column will be joined.
    default_fields_for_outer_join : dict
        When you use outer join, all output ticks will have fields from the schema of the joined source.
        If nothing was joined to a particular output tick, these fields will have default values for their type.
        This parameter allows to override the values that would be added to ticks for which nothing was joined.
        Dictionary keys should be field names, and dictionary values should be constants
        or :class:`Operation` expressions

    Returns
    -------
    :class:`Source`
            Source with joined ticks from ``collection_name``

    See also
    --------
    **JOIN_WITH_COLLECTION_SUMMARY** OneTick event processor

    Examples
    --------
    >>> # OTdirective: snippet-name: Special functions.join with collection.without query;
    >>> src = otp.Tick(A=1)
    >>> src.state_vars['TICK_SET'] = otp.state.tick_set('LATEST_TICK', 'B', otp.eval(otp.Tick(B=1, C='STR')))
    >>> src = src.join_with_collection('TICK_SET')
    >>> otp.run(src)[["A", "B", "C"]]
        A  B    C
    0  1  1  STR

    >>> # OTdirective: snippet-name: Special functions.join with collection.with query and params;
    >>> src = otp.Ticks(A=[1, 2, 3, 4, 5],
    ...                 B=[2, 2, 3, 3, 3])
    >>> src.state_vars['TICK_LIST'] = otp.state.tick_list()
    >>> def fun(tick): tick.state_vars['TICK_LIST'].push_back(tick)
    >>> src = src.script(fun)
    >>>
    >>> def join_fun(source, param_b):
    ...     source = source.agg(dict(VALUE=otp.agg.sum(source['A'])))
    ...     source['VALUE'] = source['VALUE'] + param_b
    ...     return source
    >>>
    >>> src = src.join_with_collection('TICK_LIST', join_fun, params=dict(param_b=src['B']))
    >>> otp.run(src)[["A", "B", "VALUE"]]
        A  B  VALUE
    0  1  2      3
    1  2  2      5
    2  3  3      9
    3  4  3     13
    4  5  3     18

    Join last standing quote from each exchange to trades:

    >>> # OTdirective: snippet-name: Special functions.join with collection.standing quotes per exchange;
    >>> trd = otp.Ticks(offset=[1000, 2000, 3000, 4000, 5000],
    ...                 PRICE=[10.1, 10.2, 10.15, 10.23, 10.4],
    ...                 SIZE=[100, 50, 100, 60, 200])
    >>>
    >>> qte = otp.Ticks(offset=[500, 600, 1200, 2500, 3500, 3600, 4800],
    ...                 EXCHANGE=['N', 'C', 'Q', 'Q', 'C', 'N', 'C'],
    ...                 ASK_PRICE=[10.2, 10.18, 10.18, 10.15, 10.31, 10.32, 10.44],
    ...                 BID_PRICE=[10.1, 10.17, 10.17, 10.1, 10.23, 10.31, 10.4])
    >>>
    >>> trd['TICK_TYPE'] = 'TRD'
    >>> qte['TICK_TYPE'] = 'QTE'
    >>>
    >>> trd_qte = trd + qte
    >>> trd_qte.state_vars['LAST_QUOTE_PER_EXCHANGE'] = otp.state.tick_set(
    ...     'LATEST', 'EXCHANGE',
    ...     schema=['EXCHANGE', 'ASK_PRICE', 'BID_PRICE'])
    >>>
    >>> trd_qte = trd_qte.state_vars['LAST_QUOTE_PER_EXCHANGE'].update(where=trd_qte['TICK_TYPE'] == 'QTE',
    ...                                                                value_fields=['ASK_PRICE', 'BID_PRICE'])
    >>> trd, _ = trd_qte[trd_qte['TICK_TYPE'] == 'TRD']
    >>> trd.drop(['ASK_PRICE', 'BID_PRICE', 'EXCHANGE'], inplace=True)
    >>> trd = trd.join_with_collection('LAST_QUOTE_PER_EXCHANGE')
    >>> otp.run(trd)[['PRICE', 'SIZE', 'EXCHANGE', 'ASK_PRICE', 'BID_PRICE']]
        PRICE  SIZE EXCHANGE  ASK_PRICE  BID_PRICE
    0   10.10   100        N      10.20      10.10
    1   10.10   100        C      10.18      10.17
    2   10.20    50        N      10.20      10.10
    3   10.20    50        C      10.18      10.17
    4   10.20    50        Q      10.18      10.17
    5   10.15   100        N      10.20      10.10
    6   10.15   100        C      10.18      10.17
    7   10.15   100        Q      10.15      10.10
    8   10.23    60        N      10.32      10.31
    9   10.23    60        C      10.31      10.23
    10  10.23    60        Q      10.15      10.10
    11  10.40   200        N      10.32      10.31
    12  10.40   200        C      10.44      10.40
    13  10.40   200        Q      10.15      10.10
    """

    # check that passed collection is good
    if collection_name not in self.state_vars.names:
        raise KeyError(f'Collection with name {collection_name} is not in the list of available state variables')

    if not isinstance(self.state_vars[collection_name], _TickSequence):
        raise ValueError(
            f'State variable {collection_name} is not a tick collection! '
            'Only TickSet, TickSetUnordered, TickList and TickDeque objects are supported '
            'as data sources for join_with_collection'
        )

    if params is None:
        params = {}

    special_params = ('source', '__fixed_start_time', '__fixed_end_time')
    for sp_param in special_params:
        if sp_param in params.keys():
            raise ValueError(
                f'Parameter name "{sp_param}" is special and cannot be used for params '
                'of join_with_collection function. Please, select a different name.'
            )

    # JOIN_WITH_COLLECTION_SUMMARY has START_TIME and END_TIME parameters with the precision of millisecond.
    # So, here we add a workaround on onetick.py side to support nsectime precision
    # "start" and "end" parameters of the EP are kept as they may be necessary
    # for performance reasons

    if start is not None:
        params['__fixed_start_time'] = start
        start = start - otp.Milli(1)

    if end is not None:
        params['__fixed_end_time'] = end
        end = end + otp.Milli(1)

    # prepare temporary file
    # ------------------------------------ #

    # TODO: this should be a common code somewhere
    collection_schema = {
        key: value
        for key, value in self.state_vars[collection_name].schema.items()
        if not self._check_key_is_reserved(key)
    }

    join_source_root = otp.DataSource(
        db=otp.config.default_db, tick_type="ANY", schema_policy="manual", schema=collection_schema,
    )
    if query_func is None:
        query_func = lambda source: source  # noqa

    converted_params = prepare_params(**params)

    fixed_start_time = None
    fixed_end_time = None
    if '__fixed_start_time' in converted_params.keys():
        fixed_start_time = converted_params['__fixed_start_time']
        del converted_params['__fixed_start_time']
    if '__fixed_end_time' in converted_params.keys():
        fixed_end_time = converted_params['__fixed_end_time']
        del converted_params['__fixed_end_time']

    sub_source = query_func(source=join_source_root, **converted_params)

    if fixed_start_time is not None:
        sub_source = sub_source[sub_source['TIMESTAMP'] >= fixed_start_time][0]
    if fixed_end_time is not None:
        sub_source = sub_source[sub_source['TIMESTAMP'] < fixed_end_time][0]

    sub_source = self._process_keep_time_param(keep_time, sub_source)

    params_str = _columns_to_params_for_joins(params, query_params=True)

    sub_source_schema = sub_source.schema.copy()

    columns = {}
    columns.update(self._get_columns_with_prefix(sub_source, prefix))
    columns.update(self.columns(skip_meta_fields=True))

    res = self.copy(columns=columns)

    res._merge_tmp_otq(sub_source)
    query_name = sub_source._store_in_tmp_otq(
        res._tmp_otq, symbols='_NON_EXISTING_SYMBOL_', operation_suffix="join_with_collection"
    )
    # ------------------------------------ #
    default_fields_for_outer_join_str = _get_default_fields_for_outer_join_str(
        default_fields_for_outer_join, how, sub_source_schema
    )

    join_params = dict(
        collection_name=str(self.state_vars[collection_name]),
        otq_query=f'"THIS::{query_name}"',
        join_type=how.upper(),
        otq_query_params=params_str,
        default_fields_for_outer_join=default_fields_for_outer_join_str,
    )

    _fill_aux_params_for_joins(
        join_params, caching, end, prefix, start, symbol_name=None, timezone=None, for_join_with_collection=True
    )
    res.sink(otq.JoinWithCollectionSummary(**join_params))
    res._add_table()
    res.sink(otq.Passthrough(fields="TIMESTAMP", drop_fields=True))

    return res


def join_with_query(
    self: 'Source',
    query,
    how="outer",
    symbol=None,
    params=None,
    start=None,
    end=None,
    timezone=None,
    prefix=None,
    caching=None,
    keep_time=None,
    where=None,
    default_fields_for_outer_join=None,
    symbol_time=None,
    concurrency=None,
    process_query_async: bool = True,
    **kwargs,
) -> 'Source':
    """
    For each tick executes ``query``.

    Parameters
    ----------
    query: callable, Source
        Callable ``query`` should return :class:`Source`. This object will be evaluated by OneTick (not python)
        for every tick. Note python code will be executed only once, so all python's conditional expressions
        will be evaluated only once too.
        Callable should have ``symbol`` parameter and the parameters with names
        from ``params`` if they are specified in this method.

        If ``query`` is a :class:`Source` object then it will be propagated as a query to OneTick.
    how: 'inner', 'outer'
        Type of join. If **inner**, then each tick is propagated
        only if its ``query`` execution has a non-empty result.
    params: dict
        Mapping of the parameters' names and their values for the ``query``.
        :py:class:`Columns <onetick.py.Column>` can be used as a value.
    symbol: str, Operation, dict, Source, or Tuple[Union[str, Operation], Union[dict, Source]]
        Symbol name to use in ``query``. In addition, symbol params can be passed along with symbol name.

        Symbol name can be passed as a string or as an :class:`Operation`.

        Symbol parameters can be passed as a dictionary. Also, the main :class:`Source` object,
        or the object containing a symbol parameter list, can be used as a list of symbol parameter.
        Special symbol parameters (`_PARAM_START_TIME_NANOS` and `_PARAM_END_TIME_NANOS`)
        will be ignored and will not be propagated to ``query``.

        ``symbol`` will be interpreted as a symbol name or as symbol parameters, depending on its type.
        You can pass both as a tuple.

        If symbol name is not passed, then symbol name from the main source is used.
    start: :py:class:`otp.datetime <onetick.py.datetime>`, :py:class:`otp.Operation <onetick.py.Operation>`
        Start time of ``query``.
        By default, start time of the main source is used.
    end: :py:class:`otp.datetime <onetick.py.datetime>`, :py:class:`otp.Operation <onetick.py.Operation>`
        End time of ``query`` (note that it's non-inclusive).
        By default, end time of the main source is used.
    start_time:
        .. deprecated:: 1.48.4
            The same as ``start``.
    end_time:
        .. deprecated:: 1.48.4
            The same as ``end``.
    timezone : Optional, str
        Timezone of ``query``.
        By default, timezone of the main source is used.
    prefix : str
        Prefix for the names of joined tick fields.
    caching : str
        If `None` caching is disabled (default). You can specify caching by using values:

            * 'cross_symbol': cache is the same for all symbols

            * 'per_symbol': cache is different for each symbol.

        .. note::
            When parameter ``process_query_async`` is set to ``True`` (default), caching may work
            unexpectedly, because ticks will be accumulated in batches and ``query`` will be processed
            in different threads.
    keep_time : str
        Name for the joined timestamp column. `None` means no timestamp column will be joined.
    where : Operation
        Condition to filter ticks for which the result of the ``query`` will be joined.
    default_fields_for_outer_join : dict
        When you use outer join, all output ticks will have fields from the schema of the joined source.
        If nothing was joined to a particular output tick, these fields will have default values for their type.
        This parameter allows to override the values that would be added to ticks for which nothing was joined.
        Dictionary keys should be field names, and dictionary values should be constants
        or :class:`Operation` expressions
    symbol_time : :py:class:`otp.datetime <onetick.py.datetime>`, :py:class:`otp.Operation <onetick.py.Operation>`
        Time that will be used by Onetick to map the symbol with which ``query`` is executed to the reference data.
        This parameter is only necessary if the query is expected to perform symbology conversions.
    concurrency : int
        Specifies number of threads for asynchronous processing of ``query`` per unbound symbol list.
        By default, the number of threads is 1.
    process_query_async: bool
        Switches between synchronous and asynchronous execution of queries.

        While asynchronous execution is generally much more effective,
        in certain cases synchronous execution may still be preferred
        (e.g., when there are a few input ticks, each initiating a memory-consuming query).

        In asynchronous mode typically while parallel thread is processing the query,
        EP accumulates some input ticks.


    Returns
    -------
    :class:`Source`
            Source with joined ticks from ``query``

    See also
    --------
    **JOIN_WITH_QUERY** OneTick event processor

    Examples
    --------
    >>> # OTdirective: snippet-name: Special functions.join with query.with an otp data source;
    >>> d = otp.Ticks(Y=[-1])
    >>> d = d.update(dict(Y=1), where=(d.Symbol.name == "a"))
    >>> data = otp.Ticks(X=[1, 2],
    ...                  S=["a", "b"])
    >>> res = data.join_with_query(d, how='inner', symbol=data['S'])
    >>> otp.run(res)[["X", "Y", "S"]]
       X  Y  S
    0  1  1  a
    1  2 -1  b

    >>> d = otp.Ticks(ADDED=[-1])
    >>> d = d.update(dict(ADDED=1), where=(d.Symbol.name == "3"))  # symbol name is always string
    >>> data = otp.Ticks(A=[1, 2], B=[2, 4])
    >>> res = data.join_with_query(d, how='inner', symbol=(data['A'] + data['B']))  # OTdirective: skip-snippet:;
    >>> df = otp.run(res)
    >>> df[["A", "B", "ADDED"]]
       A  B  ADDED
    0  1  2      1
    1  2  4     -1

    Constants as symbols are also supported:

    >>> d = otp.Ticks(ADDED=[d.Symbol.name])
    >>> data = otp.Ticks(A=[1, 2], B=[2, 4])
    >>> res = data.join_with_query(d, how='inner', symbol=1)    # OTdirective: skip-snippet:;
    >>> df = otp.run(res)
    >>> df[["A", "B", "ADDED"]]
       A  B ADDED
    0  1  2     1
    1  2  4     1

    Function object as query is also supported (Note it will be executed only once in python's code):

    >>> def func(symbol):
    ...    d = otp.Ticks(TYPE=["six"])
    ...    d = d.update(dict(TYPE="three"), where=(symbol.name == "3"))  # symbol is always converted to string
    ...    d["TYPE"] = symbol['PREF'] + d["TYPE"] + symbol['POST']
    ...    return d
    >>> # OTdirective: snippet-name: Special functions.join with query.with a function
    >>> data = otp.Ticks(A=[1, 2], B=[2, 4])
    >>> res = data.join_with_query(func, how='inner', symbol=(data['A'] + data['B'], dict(PREF="_", POST="$")))
    >>> df = otp.run(res)
    >>> df[["A", "B", "TYPE"]]
       A  B     TYPE
    0  1  2  _three$
    1  2  4    _six$

    It's possible to pass the source itself as a list of symbol parameters, which will make all of its fields
    accessible through the "symbol" object:

    >>> def func(symbol):
    ...    d = otp.Ticks(TYPE=["six"])
    ...    d["TYPE"] = symbol['PREF'] + d["TYPE"] + symbol['POST']
    ...    return d
    >>> # OTdirective: snippet-name: 'Source' operations.join with query.source as symbol;
    >>> data = otp.Ticks(A=[1, 2], B=[2, 4], PREF=["_", "$"], POST=["$", "_"])
    >>> res = data.join_with_query(func, how='inner', symbol=data)
    >>> df = otp.run(res)
    >>> df[["A", "B", "TYPE"]]
       A  B   TYPE
    0  1  2  _six$
    1  2  4  $six_

    The examples above can be rewritten by using onetick query parameters instead of symbol parameters.
    OTQ parameters are global for query, while symbol parameters can be redefined by bound symbols:

    >>> def func(symbol, pref, post):
    ...     d = otp.Ticks(TYPE=["six"])
    ...     d = d.update(dict(TYPE="three"), where=(symbol.name == "3"))  # symbol is always converted to string
    ...     d["TYPE"] = pref + d["TYPE"] + post
    ...     return d
    >>> # OTdirective: snippet-name: Special functions.join with query.with a function that takes params;
    >>> data = otp.Ticks(A=[1, 2], B=[2, 4])
    >>> res = data.join_with_query(func, how='inner', symbol=(data['A'] + data['B']),
    ...                            params=dict(pref="_", post="$"))
    >>> df = otp.run(res)
    >>> df[["A", "B", "TYPE"]]
       A  B     TYPE
    0  1  2  _three$
    1  2  4    _six$

    Some or all onetick query parameters can be column or expression also:

    >>> def func(symbol, pref, post):
    ...     d = otp.Ticks(TYPE=["six"])
    ...     d = d.update(dict(TYPE="three"), where=(symbol.name == "3"))  # symbol is always converted to string
    ...     d["TYPE"] = pref + d["TYPE"] + post
    ...     return d
    >>> # OTdirective: snippet-name: Special functions.join with query.with a function that takes params from fields;   # noqa
    >>> data = otp.Ticks(A=[1, 2], B=[2, 4], PREF=["^", "_"], POST=["!", "$"])
    >>> res = data.join_with_query(func, how='inner', symbol=(data['A'] + data['B']),
    ...                            params=dict(pref=data["PREF"] + ".", post=data["POST"]))
    >>> df = otp.run(res)
    >>> df[["A", "B", "TYPE"]]
       A  B      TYPE
    0  1  2  ^.three!
    1  2  4    _.six$

    You can specify ``start`` and ``end`` time of the query, otherwise time interval of the main query will be used:

    >>> # OTdirective: snippet-name: Special functions.join with query.passing start/end times;
    >>> d = otp.Ticks(Y=[1, 2])
    >>> data = otp.Ticks(X=[1, 2])
    >>> start = otp.datetime(2003, 12, 1, 0, 0, 0, 1000)
    >>> end = otp.datetime(2003, 12, 1, 0, 0, 0, 3000)
    >>> res = data.join_with_query(d, how='inner', start=start, end=end)
    >>> otp.run(res)
                         Time  Y  X
    0 2003-12-01 00:00:00.000  1  1
    1 2003-12-01 00:00:00.000  2  1
    2 2003-12-01 00:00:00.001  1  2
    3 2003-12-01 00:00:00.001  2  2

    By default joined query inherits start and end time from the main query:

    >>> joined_query = otp.Tick(JOINED_START_TIME=otp.meta_fields.start_time,
    ...                         JOINED_END_TIME=otp.meta_fields.end_time)
    >>> main_query = otp.Tick(A=1)
    >>> data = main_query.join_with_query(joined_query)
    >>> otp.run(data, start=otp.dt(2003, 12, 1), end=otp.dt(2003, 12, 4))
            Time JOINED_START_TIME JOINED_END_TIME  A
    0 2003-12-01        2003-12-01      2003-12-04  1

    Parameters ``start`` and ``end`` can be used to change time interval for the joined query:

    >>> data = main_query.join_with_query(joined_query, start=otp.dt(2024, 1, 1), end=otp.dt(2024, 1, 3))
    >>> otp.run(data, start=otp.dt(2003, 12, 1), end=otp.dt(2003, 12, 4))
            Time JOINED_START_TIME JOINED_END_TIME  A
    0 2003-12-01        2024-01-01      2024-01-03  1

    Note that query ``start`` time is inclusive, but query ``end`` time is not,
    meaning that ticks with timestamps equal to the query end time will not be included:

    >>> main_query = otp.Tick(A=1)
    >>> joined_query = otp.Tick(DAY=0, bucket_interval=24*60*60)
    >>> joined_query['DAY'] = joined_query['TIMESTAMP'].dt.day_of_month()
    >>> otp.run(joined_query, start=otp.dt(2003, 12, 1), end=otp.dt(2003, 12, 5))
            Time  DAY
    0 2003-12-01    1
    1 2003-12-02    2
    2 2003-12-03    3
    3 2003-12-04    4

    >>> joined_query = joined_query.last()
    >>> data = main_query.join_with_query(joined_query,
    ...                                   start=otp.dt(2003, 12, 1), end=otp.dt(2003, 12, 4))
    >>> otp.run(data)
            Time  DAY  A
    0 2003-12-01    3  1

    If you want to include such ticks, you can add one nanosecond to the query end time:

    >>> data = main_query.join_with_query(joined_query,
    ...                                   start=otp.dt(2003, 12, 1), end=otp.dt(2003, 12, 4) + otp.Nano(1))
    >>> otp.run(data)
            Time  DAY  A
    0 2003-12-01    4  1

    Use ``keep_time`` parameter to keep or rename original timestamp column:

    >>> # OTdirective: snippet-name: Special functions.join with query.keep the timestamps of the joined ticks;
    >>> d = otp.Ticks(Y=[1, 2])
    >>> data = otp.Ticks(X=[1, 2])
    >>> res = data.join_with_query(d, how='inner', keep_time="ORIG_TIME")
    >>> otp.run(res)
                         Time  Y               ORIG_TIME  X
    0 2003-12-01 00:00:00.000  1 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.000  2 2003-12-01 00:00:00.001  1
    2 2003-12-01 00:00:00.001  1 2003-12-01 00:00:00.000  2
    3 2003-12-01 00:00:00.001  2 2003-12-01 00:00:00.001  2
    """

    # TODO: check if join_with_query checks schema of joined source against primary source,
    # by itself or with process_by_group

    if params is None:
        params = {}

    converted_symbol_name, symbol_param = _check_and_convert_symbol(symbol)

    # default symbol name should be this: _SYMBOL_NAME if it is not empty else _NON_EXISTING_SYMBOL_
    # this way we will force JWQ to substitute symbol with any symbol parameters we may have passed
    # otherwise (if an empty symbol name is passed to JWQ), it will not substitute either symbol name
    # or symbol parameters, and so symbol parameters may get lost
    # see BDS-263
    if converted_symbol_name is None:
        converted_symbol_name = "CASE(_SYMBOL_NAME,'','_NON_EXISTING_SYMBOL',_SYMBOL_NAME)"

    converted_symbol_param_columns, converted_symbol_param = _convert_symbol_param_and_columns(symbol_param)
    if converted_symbol_param is None:
        # we couldn't interpret "symbols" as either symbol name or symbol parameters
        raise ValueError(
            '"symbol" parameter has a wrong format! It should be a symbol name, a symbol parameter '
            'object (dict or Source), or a tuple containing both'
        )

    # adding symbol time
    if '_PARAM_SYMBOL_TIME' in converted_symbol_param_columns.keys():
        warnings.warn(
            '"_PARAM_SYMBOL_TIME" explicitly passed among join_with_query symbol parameters! '
            'This is deprecated - please use symbol_time parameter instead. '
            'If you specify symbol_time parameter, it will override the explicitly passed value',
            FutureWarning,
            stacklevel=2,
        )
    if symbol_time is not None:
        if ott.get_object_type(symbol_time) is not otp.nsectime and ott.get_object_type(symbol_time) is not str:
            raise ValueError(
                f'Parameter of type {ott.get_object_type(symbol_time)} passed as symbol_time! '
                'This parameter only supports datetime values or strings'
            )
        converted_symbol_param_columns['_PARAM_SYMBOL_TIME'] = symbol_time

    # prepare temporary file
    # ------------------------------------ #
    converted_params = prepare_params(**params)

    if isinstance(query, otp.Source):
        sub_source = query
    else:
        # inspect function
        # -------
        sig = inspect.signature(query)
        if "symbol" in sig.parameters:
            if "symbol" in converted_params.keys():
                raise AttributeError(
                    '"params" contains key "symbol", which is reserved for symbol parameters. '
                    'Please, rename this parameter to another name'
                )
            converted_params["symbol"] = converted_symbol_param  # type: ignore
        sub_source = query(**converted_params)

    sub_source = self._process_keep_time_param(keep_time, sub_source)

    if not sub_source._is_unbound_required():
        sub_source += otp.Empty()

    params_str = _columns_to_params_for_joins(params, query_params=True)
    symbol_params_str = _columns_to_params_for_joins(converted_symbol_param_columns)

    sub_source_schema = sub_source.schema.copy()

    columns = {}
    columns.update(self._get_columns_with_prefix(sub_source, prefix))
    columns.update(self.columns(skip_meta_fields=True))

    res = self.copy(columns=columns)

    res._merge_tmp_otq(sub_source)
    query_name = sub_source._store_in_tmp_otq(
        res._tmp_otq, symbols='_NON_EXISTING_SYMBOL_', operation_suffix="join_with_query"
    )  # TODO: combine with _convert_symbol_to_string
    # ------------------------------------ #

    if where is not None and how != 'outer':
        raise ValueError('The `where` parameter can be used only for outer join')

    default_fields_for_outer_join_str = _get_default_fields_for_outer_join_str(
        default_fields_for_outer_join, how, sub_source_schema
    )

    join_params = dict(
        otq_query=f'"THIS::{query_name}"',
        join_type=how.upper(),
        otq_query_params=params_str,
        symbol_params=symbol_params_str,
        where=str(where._make_python_way_bool_expression()) if where is not None else '',
        default_fields_for_outer_join=default_fields_for_outer_join_str,
        process_query_asynchronously=process_query_async,
    )
    if concurrency is not None:
        if not isinstance(concurrency, int) or concurrency <= 0:
            raise ValueError('Wrong value of concurrency parameter passed! concurrency should be a positive integer')
        join_params['shared_thread_count'] = concurrency

    start_time = kwargs.get('start_time', start)
    end_time = kwargs.get('end_time', end)
    _fill_aux_params_for_joins(join_params, caching, end_time, prefix, start_time, converted_symbol_name, timezone)
    res.sink(otq.JoinWithQuery(**join_params))
    res._add_table()
    res.sink(otq.Passthrough(fields="TIMESTAMP", drop_fields=True))

    return res


def point_in_time(
    self: 'Source',
    source: Union['Source', str],
    offsets: List[int],
    offset_type: Literal['time_msec', 'num_ticks'] = 'time_msec',
    input_ts_fields_to_propagate: Optional[List[str]] = None,
    symbol_date=None,
) -> 'Source':
    """
    This method joins ticks from current source with the ticks from another ``source``.

    Joined ticks are those that are offset by
    the specified number of milliseconds or by the specified number of ticks
    relative to the current source's tick timestamp.

    Output tick may be generated for each specified offset, so this method may output several ticks for each input tick.

    If another ``source`` doesn't have a tick with specified offset, then output tick is not generated.

    Fields **TICK_TIME** and **OFFSET** are also added to the output ticks,
    specifying original timestamp of the joined tick and the offset that it was specified to join by.

    Note
    ----
    In order for this method to have reasonable performance,
    the set of input ticks' timestamps has to be relatively small.

    In other words, the points in time, which the user is interested in,
    have to be quite few in order usage of this method to be justified.

    Parameters
    ----------
    source: :class:`Source` or str
        The source from which the data will be joined or the string with the path to the .otq file
        (note that in the latter case schema can not be updated automatically with the fields from the joined query).
    offsets:
        List of integers specifying offsets for each timestamp.
    offset_type: 'time_msec' or 'num_ticks'
        The type of offset: number of milliseconds or the number of ticks.
    input_ts_fields_to_propagate:
        The list of fields to propagate from the current source.
        By default no fields (except **TIMESTAMP**) are propagated.
    symbol_date: :py:class:`otp.datetime <onetick.py.datetime>`
        Symbol date that will be set for the ``source`` inner query.

    See also
    --------
    | **POINT_IN_TIME** OneTick event processor
    | :func:`onetick.py.PointInTime`
    | :func:`onetick.py.join_by_time`

    Examples
    --------

    Quotes and trades for testing:

    .. testcode::

       qte = otp.Ticks(ASK_PRICE=[20, 21, 22, 23, 24, 25], BID_PRICE=[20, 21, 22, 23, 24, 25])
       print(otp.run(qte))

    .. testoutput::

                            Time  ASK_PRICE  BID_PRICE
       0 2003-12-01 00:00:00.000         20         20
       1 2003-12-01 00:00:00.001         21         21
       2 2003-12-01 00:00:00.002         22         22
       3 2003-12-01 00:00:00.003         23         23
       4 2003-12-01 00:00:00.004         24         24
       5 2003-12-01 00:00:00.005         25         25

    .. testcode::

       trd = otp.Ticks(PRICE=[1, 3, 5], SIZE=[100, 300, 500], offset=[1, 3, 5])
       print(otp.run(trd))

    .. testoutput::

                            Time  PRICE  SIZE
       0 2003-12-01 00:00:00.001      1   100
       1 2003-12-01 00:00:00.003      3   300
       2 2003-12-01 00:00:00.005      5   500

    Joining each quote with first trade with equal or less timestamp:

    .. testcode::
       :skipif: not is_supported_point_in_time()

       data = qte.point_in_time(trd, offsets=[0])
       print(otp.run(data))

    .. testoutput::

                            Time  PRICE  SIZE               TICK_TIME  OFFSET
       0 2003-12-01 00:00:00.001      1   100 2003-12-01 00:00:00.001       0
       1 2003-12-01 00:00:00.002      1   100 2003-12-01 00:00:00.001       0
       2 2003-12-01 00:00:00.003      3   300 2003-12-01 00:00:00.003       0
       3 2003-12-01 00:00:00.004      3   300 2003-12-01 00:00:00.003       0
       4 2003-12-01 00:00:00.005      5   500 2003-12-01 00:00:00.005       0

    By default fields from the current source are not propagated,
    use parameter ``input_ts_fields_to_propagate`` to add them to the output:

    .. testcode::
       :skipif: not is_supported_point_in_time()

       data = qte.point_in_time(trd, offsets=[0], input_ts_fields_to_propagate=['ASK_PRICE', 'BID_PRICE'])
       print(otp.run(data))

    .. testoutput::

                            Time  ASK_PRICE  BID_PRICE  PRICE  SIZE               TICK_TIME  OFFSET
       0 2003-12-01 00:00:00.001         21         21      1   100 2003-12-01 00:00:00.001       0
       1 2003-12-01 00:00:00.002         22         22      1   100 2003-12-01 00:00:00.001       0
       2 2003-12-01 00:00:00.003         23         23      3   300 2003-12-01 00:00:00.003       0
       3 2003-12-01 00:00:00.004         24         24      3   300 2003-12-01 00:00:00.003       0
       4 2003-12-01 00:00:00.005         25         25      5   500 2003-12-01 00:00:00.005       0

    Note that first quote was not propagated, because it doesn't have corresponding trade.

    Offset may be positive or negative.
    If several offsets are specified, several output ticks may be generated for a single input tick:

    .. testcode::
       :skipif: not is_supported_point_in_time()

       data = qte.point_in_time(trd, offsets=[0, 1], input_ts_fields_to_propagate=['ASK_PRICE', 'BID_PRICE'])
       print(otp.run(data))

    .. testoutput::

                             Time  ASK_PRICE  BID_PRICE  PRICE  SIZE               TICK_TIME  OFFSET
       0  2003-12-01 00:00:00.000         20         20      1   100 2003-12-01 00:00:00.001       1
       1  2003-12-01 00:00:00.001         21         21      1   100 2003-12-01 00:00:00.001       0
       2  2003-12-01 00:00:00.001         21         21      1   100 2003-12-01 00:00:00.001       1
       3  2003-12-01 00:00:00.002         22         22      1   100 2003-12-01 00:00:00.001       0
       4  2003-12-01 00:00:00.002         22         22      3   300 2003-12-01 00:00:00.003       1
       5  2003-12-01 00:00:00.003         23         23      3   300 2003-12-01 00:00:00.003       0
       6  2003-12-01 00:00:00.003         23         23      3   300 2003-12-01 00:00:00.003       1
       7  2003-12-01 00:00:00.004         24         24      3   300 2003-12-01 00:00:00.003       0
       8  2003-12-01 00:00:00.004         24         24      5   500 2003-12-01 00:00:00.005       1
       9  2003-12-01 00:00:00.005         25         25      5   500 2003-12-01 00:00:00.005       0
       10 2003-12-01 00:00:00.005         25         25      5   500 2003-12-01 00:00:00.005       1

    By default the number of milliseconds is used as an offset.
    You can also specify the number of ticks as an offset:

    .. testcode::
       :skipif: not is_supported_point_in_time()

       data = qte.point_in_time(trd, offset_type='num_ticks', offsets=[-1, 1],
                                input_ts_fields_to_propagate=['ASK_PRICE', 'BID_PRICE'])
       print(otp.run(data))

    .. testoutput::

                            Time  ASK_PRICE  BID_PRICE  PRICE  SIZE               TICK_TIME  OFFSET
       0 2003-12-01 00:00:00.000         20         20      1   100 2003-12-01 00:00:00.001       1
       1 2003-12-01 00:00:00.001         21         21      3   300 2003-12-01 00:00:00.003       1
       2 2003-12-01 00:00:00.002         22         22      3   300 2003-12-01 00:00:00.003       1
       3 2003-12-01 00:00:00.003         23         23      1   100 2003-12-01 00:00:00.001      -1
       4 2003-12-01 00:00:00.003         23         23      5   500 2003-12-01 00:00:00.005       1
       5 2003-12-01 00:00:00.004         24         24      1   100 2003-12-01 00:00:00.001      -1
       6 2003-12-01 00:00:00.004         24         24      5   500 2003-12-01 00:00:00.005       1
       7 2003-12-01 00:00:00.005         25         25      3   300 2003-12-01 00:00:00.003      -1
    """
    if not is_supported_point_in_time():
        raise RuntimeError('PointInTime event processor is not supported on this OneTick version')

    res = self.copy()

    if offset_type not in ('time_msec', 'num_ticks'):
        raise ValueError(f"Wrong value for parameter 'offset_type': {offset_type}")

    if isinstance(source, str):
        otq_query = source
    else:
        query_name = source._store_in_tmp_otq(
            res._tmp_otq,
            operation_suffix='point_in_time',
            # set default symbol, even if it's not set by user, symbol's value doesn't matter in this case
            symbols=otp.config.get('default_symbol', 'ANY'),
            symbol_date=symbol_date,
        )
        otq_query = f'THIS::{query_name}'

    input_ts_fields_to_propagate = input_ts_fields_to_propagate or []

    pit_params = dict(
        otq_query=otq_query,
        offset_type=offset_type.upper(),
        offsets=','.join(map(str, offsets)),
        input_ts_fields_to_propagate=','.join(map(str, input_ts_fields_to_propagate)),
    )
    res.sink(otq.PointInTime(**pit_params))

    schema = {}
    if input_ts_fields_to_propagate:
        schema = {
            k: v for k, v in res.schema.items()
            if k in input_ts_fields_to_propagate
        }
    res.schema.set(**schema)
    if not isinstance(source, str):
        res.schema.update(**source.schema)
    res.schema.update(**{
        'TICK_TIME': otp.nsectime,
        'OFFSET': int,
    })
    return res


def join_with_snapshot(
    self: 'Source',
    snapshot_name='VALUE',
    snapshot_storage='memory',
    allow_snapshot_absence=False,
    join_keys=None,
    symbol_name_in_snapshot=None,
    database='',
    default_fields_for_outer_join=None,
    prefix_for_output_ticks='',
    snapshot_fields=None,
):
    """
    Saves last (at most) `n` ticks of each group of ticks from the input time series in global storage or
    in a memory mapped file under a specified snapshot name.
    Tick descriptor should be the same for all ticks saved into the snapshot.
    These ticks can then be read via :py:class:`ReadSnapshot <onetick.py.ReadSnapshot>` by using the name
    of the snapshot and the same symbol name (``<db_name>::<symbol>``) that were used by this method.

    .. warning::
        You should update schema manually, if you want to use fields from snapshot in `onetick-py` query description
        before its execution.

        That's due to the fact, that `onetick-py` can't identify a schema of data in a snapshot before making a query.

        If you set ``default_fields_for_outer_join`` parameter, schema will be guessed from default fields values.

    Parameters
    ----------
    snapshot_name: str
        The name that was specified in :py:meth:`onetick.py.Source.save_snapshot` as a ``snapshot_name`` during saving.

        Default: `VALUE`
    snapshot_storage: str
        This parameter specifies the place of storage of the snapshot. Possible options are:

        * `memory` - the snapshot is stored in the dynamic (heap) memory of the process
          that ran (or is still running) the :py:meth:`onetick.py.Source.save_snapshot` for the snapshot.
        * `memory_mapped_file` - the snapshot is stored in a memory mapped file.
          For each symbol to get the location of the snapshot in the file system, ``join_with_snapshot`` looks at
          the **SAVE_SNAPSHOT_DIR** parameter value in the locator section for the database of the symbol.
          In a specified directory it creates a new directory with the name of the snapshot and keeps
          the memory mapped file and some other helper files there.

        Default: `memory`
    allow_snapshot_absence: bool
        If specified, the EP does not display an error about missing snapshot
        if the snapshot has not been saved or is still being saved.

        Default: `False`
    join_keys: list, optional
        A list of names of attributes. A non-empty list causes input ticks to be joined only if all of them
        have matching values for all specified attributes.
        Currently, these fields need to match with ``group_by`` fields of the corresponding snapshot.
    symbol_name_in_snapshot: str, :class:`~onetick.py.Column` or :class:`~onetick.py.Operation`, optional
        Expression that evaluates to a string containing symbol name.
        Specified expression is reevaluated upon the arrival of each tick.
        If this parameter is empty, the input symbol name is used.
    database: str, optional
        The database to read the snapshot. If not specified database from the symbol is used.
    default_fields_for_outer_join: dict, optional
        A `dict` with field name as key and value, :class:`~onetick.py.Column` or :class:`~onetick.py.Operation`,
        which specifies the names and the values of the fields (also, optionally, the field type),
        used to form ticks to be joined with unmatched input ticks.

        If you want to specify field type, pass tuple of field dtype and expression or value as dict item value.

        This parameter is reevaluated upon the arrival of each tick.

        It's also used for auto detecting snapshot schema for using fields from snapshot
        while building query via ``ontick-py``.
    prefix_for_output_ticks: str
        The prefix for the names of joined tick fields.

        Default: `empty string`
    snapshot_fields: List[str], None
        Specifies list of fields from the snapshot to join with input ticks. When empty, all fields are included.

    See also
    --------
    | **JOIN_WITH_SNAPSHOT** OneTick event processor
    | :py:class:`onetick.py.ReadSnapshot`
    | :py:class:`onetick.py.ShowSnapshotList`
    | :py:class:`onetick.py.FindSnapshotSymbols`
    | :py:meth:`onetick.py.Source.save_snapshot`

    Examples
    --------
    Simple ticks join with snapshot:

    >>> src = otp.Ticks(A=[1, 2])
    >>> src = src.join_with_snapshot(snapshot_name='some_snapshot')  # doctest: +SKIP
    >>> otp.run(src)  # doctest: +SKIP
                         Time  A  X  Y               TICK_TIME
    0 2003-12-01 00:00:00.000  1  1  4 2003-12-01 00:00:00.000
    1 2003-12-01 00:00:00.000  1  2  5 2003-12-01 00:00:00.001
    2 2003-12-01 00:00:00.001  2  1  4 2003-12-01 00:00:00.000
    3 2003-12-01 00:00:00.001  2  2  5 2003-12-01 00:00:00.001

    Add prefix ``T.`` for fields from snapshot:

    >>> src = otp.Ticks(A=[1, 2])
    >>> src = src.join_with_snapshot(
    ...     snapshot_name='some_snapshot', prefix_for_output_ticks='T.',
    ... )  # doctest: +SKIP
    >>> otp.run(src)  # doctest: +SKIP
                         Time  A  T.X  T.Y             T.TICK_TIME
    0 2003-12-01 00:00:00.000  1    1    4 2003-12-01 00:00:00.000
    1 2003-12-01 00:00:00.000  1    2    5 2003-12-01 00:00:00.001
    2 2003-12-01 00:00:00.001  2    1    4 2003-12-01 00:00:00.000
    3 2003-12-01 00:00:00.001  2    2    5 2003-12-01 00:00:00.001

    To get only specific fields from snapshot use parameter ``snapshot_fields``:

    >>> src = otp.Ticks(A=[1, 2])
    >>> src = src.join_with_snapshot(
    ...     snapshot_name='some_snapshot', snapshot_fields=['Y'],
    ... )  # doctest: +SKIP
    >>> otp.run(src)  # doctest: +SKIP
                         Time  A  Y
    0 2003-12-01 00:00:00.000  1  4
    1 2003-12-01 00:00:00.000  1  5
    2 2003-12-01 00:00:00.001  2  4
    3 2003-12-01 00:00:00.001  2  5

    Setting default values for snapshot fields for outer join via ``default_fields_for_outer_join_with_types``
    parameter with example of joining ticks with absent snapshot:

    >>> src = otp.Ticks(A=[1, 2])
    >>> src = src.join_with_snapshot(
    ...     snapshot_name='some_snapshot', allow_snapshot_absence=True,
    ...     default_fields_for_outer_join={
    ...         'B': 'Some string',
    ...         'C': (float, src['A'] * 2),
    ...         'D': 50,
    ...     },
    ... )  # doctest: +SKIP
    >>> otp.run(src)  # doctest: +SKIP
                         Time  A            B    C     D
    0 2003-12-01 00:00:00.000  1  Some string  2.0  50.0
    1 2003-12-01 00:00:00.001  2  Some string  2.0  50.0

    In this case, schema for ``src`` object will be automatically detected from values for this parameter:

    >>> src.schema  # doctest: +SKIP
    {'A': <class 'int'>, 'B': <class 'str'>, 'C': <class 'float'>, 'D': <class 'int'>}


    You can join ticks from snapshot for each input tick for specified symbol name from string value or this tick
    via ``symbol_name_in_snapshot`` parameter.

    Let's create snapshot with different symbol names inside:

    >>> src = otp.Ticks(X=[1, 2, 3, 4], Y=['AAA', 'BBB', 'CCC', 'AAA'])
    >>> src = src.save_snapshot(
    ...     snapshot_name='some_snapshot', num_ticks=5, keep_snapshot_after_query=True, symbol_name_field='Y',
    ... )
    >>> otp.run(src)  # doctest: +SKIP

    Now we can join input only with ticks from snapshot with specified symbol name:

    >>> src = otp.Ticks(A=[1, 2])
    >>> src = src.join_with_snapshot(
    ...     snapshot_name='some_snapshot', symbol_name_in_snapshot='AAA',
    ... )  # doctest: +SKIP
    >>> otp.run(src)  # doctest: +SKIP
                         Time  A  X               TICK_TIME
    0 2003-12-01 00:00:00.000  1  1 2003-12-01 00:00:00.000
    1 2003-12-01 00:00:00.000  1  4 2003-12-01 00:00:00.003
    2 2003-12-01 00:00:00.001  2  1 2003-12-01 00:00:00.000
    3 2003-12-01 00:00:00.001  2  4 2003-12-01 00:00:00.003

    Or we can join each tick with ticks from snapshot with symbol name from input ticks field:

    >>> src = otp.Ticks(A=[1, 2], SYM=['AAA', 'CCC'])
    >>> src = src.join_with_snapshot(
    ...     snapshot_name='some_snapshot', symbol_name_in_snapshot=src['SYM'],
    ... )  # doctest: +SKIP
    >>> otp.run(src)  # doctest: +SKIP
                         Time  A  SYM  X               TICK_TIME
    0 2003-12-01 00:00:00.000  1  AAA  1 2003-12-01 00:00:00.000
    1 2003-12-01 00:00:00.000  1  AAA  4 2003-12-01 00:00:00.003
    2 2003-12-01 00:00:00.001  2  CCC  3 2003-12-01 00:00:00.002
    """
    kwargs = {}

    if not hasattr(otq, "JoinWithSnapshot"):
        raise RuntimeError("Current version of OneTick doesn't support JOIN_WITH_SNAPSHOT EP")

    if snapshot_storage not in ['memory', 'memory_mapped_file']:
        raise ValueError('`snapshot_storage` must be one of "memory", "memory_mapped_file"')

    is_snapshot_fields_param_supported = is_join_with_snapshot_snapshot_fields_parameter_supported()

    if snapshot_fields and not is_snapshot_fields_param_supported:
        raise RuntimeError(
            "Current version of OneTick doesn't support `snapshot_fields` parameter on JOIN_WITH_SNAPSHOT EP"
        )

    snapshot_storage = snapshot_storage.upper()

    if join_keys is None:
        join_keys_str = ''
    else:
        join_keys_str = ','.join(join_keys)

    if symbol_name_in_snapshot is None:
        symbol_name_in_snapshot = ''
    elif isinstance(symbol_name_in_snapshot, _Operation):
        symbol_name_in_snapshot = str(symbol_name_in_snapshot)

    if default_fields_for_outer_join is None:
        default_fields_for_outer_join = {}

    default_fields_list = []
    snapshot_schema = {}

    for field_name, field_value in default_fields_for_outer_join.items():
        if isinstance(field_value, tuple):
            field_type = field_value[0]

            default_fields_list.append(
                f'{field_name} {ott.type2str(field_type)} = {ott.value2str(field_value[1])}',
            )
        else:
            if isinstance(field_value, _Operation):
                field_type = field_value.dtype
            else:
                field_type = type(field_value)

            default_fields_list.append(f'{field_name} = {ott.value2str(field_value)}')

        snapshot_schema[f'{prefix_for_output_ticks}{field_name}'] = field_type

    default_fields_str = ','.join(default_fields_list)

    if snapshot_fields is not None:
        kwargs['snapshot_fields'] = ','.join(snapshot_fields)

    self.sink(
        otq.JoinWithSnapshot(
            snapshot_name=snapshot_name,
            snapshot_storage=snapshot_storage,
            allow_snapshot_absence=allow_snapshot_absence,
            join_keys=join_keys_str,
            symbol_name_in_snapshot=symbol_name_in_snapshot,
            database=database,
            default_fields_for_outer_join=default_fields_str,
            prefix_for_output_ticks=prefix_for_output_ticks,
            **kwargs,
        )
    )

    self.schema.update(**snapshot_schema)

    return self



def every_operand(op):
    operations = [op]
    while operations:
        operation = operations.pop()
        operands = getattr(operation, "_op_params", None)
        if operands:
            operations.extend(operands)
        else:
            yield operation



from collections import namedtuple

from onetick.py import types as ott
from onetick.py.core.column_operations._methods.op_types import are_strings
from onetick.py.types import value2str

MethodResult = namedtuple("MethodResult", ("op_str", "dtype"))


def _wrap_object(o):
    if isinstance(o, str) or are_strings(getattr(o, "dtype", None)):
        # In PER_TICK_SCRIPT: parenthesis are not allowed in string expressions.
        return value2str(o)
    return f"({value2str(o)})"


def _type_error_for_op(op, types):
    return TypeError(f"Unsupported operand type(s) for {op} operation: {types}")


def _init_binary_op(prev_op, other):
    left = _wrap_object(prev_op)
    right = _wrap_object(other)
    left_t = ott.get_object_type(prev_op)
    right_t = ott.get_object_type(other)
    dtype = None
    op_str = None
    return left, right, left_t, right_t, op_str, dtype



import os

from onetick.py.otq import otq, _tmp_otq_path, otli, pyomd
from pandas import Timestamp as pd_Timestamp
from onetick.py.types import datetime as otp_datetime, datetime2expr
from datetime import datetime, date
from onetick.py import utils
from onetick.py.configuration import config
from onetick.py.log import get_debug_logger


def is_datetime_type(dt):
    # TODO: move to types, add docs
    return isinstance(dt, (datetime, date, pd_Timestamp, otp_datetime, pyomd.timeval_t))


def is_millisecond_precision(dt):
    """
    dt can be a datetime object, a pd.Datetime object or an otp.dt object.
    We check if it has only millisecond precision or not.
    """
    if (dt.microsecond % 1000) != 0:
        return False
    if isinstance(dt, (pd_Timestamp, otp_datetime)):
        if dt.nanosecond != 0:
            return False
    return True


class TmpOtq:
    """
    Class that represents a storage of temporary queries
    """

    class __iter_str:
        """
        Class that produces unique strings
        """
        DEFAULT_LENGTH = 6
        FIRST_CODE = 97  # 'a'
        LAST_CODE = 122  # 'z'

        def __init__(self, length=None):
            if length is None:
                length = self.DEFAULT_LENGTH
            self.__length = length
            self.__n = 0

        def __num_to_str(self):
            s = ""
            r = self.LAST_CODE - self.FIRST_CODE + 1
            for i in range(0, self.__length):
                s = chr(self.FIRST_CODE + (self.__n // r ** i) % r) + s
            return s

        def get_str(self):
            s = self.__num_to_str()
            self.__n += 1
            return s

    name_generator = __iter_str()

    def __init__(self):
        self.queries = {}

    def add_query(self, query, suffix="", name=None, params=None):
        """
        Adds query with a unique generated name to the storage.

        Parameters
        ----------
        query: otq.GraphQuery
            GraphQuery object that is being stored
        suffix: str
            Suffix that is added to the autogenerated string to form name of the query.
        name: str, optional
            If specified, this ``name`` will be used to save query
            and ``suffix`` parameter will be ignored.
        params: dict, optional
            Can specify additional parameters with which this query will be saved to file.
            Currently, only "running_query_flag" and "symbol_date" are supported

        Returns
        -------
            result: str , name of the query in the file (without THIS:: prefix).
        """
        name = name or self.name_generator.get_str() + suffix
        if name in self.queries:
            raise ValueError(f"There is already a query with name '{name}' in {self.__class__.__name__} storage")
        if params is None:
            params = {}
        self.queries[name] = (query, params)
        return name

    def merge(self, tmp_otq):
        """
        Adds queries from the tmp_otq storage to the current storage.
        As query names are guaranteed to be unique session-wide, no check for collision is necessary.
        Queries with same names will always be the same.
        """
        self.queries.update(tmp_otq.queries)

    def copy(self):
        """
        Creates a copy of the storage
        """
        res = TmpOtq()
        res.merge(self)
        return res

    def save_to_file(self, query=None, query_name="main_query", file_path=None, file_suffix="",
                     start=None, end=None, start_time_expression=None, end_time_expression=None, timezone=None,
                     running_query_flag=None,
                     symbol_date=None):
        """
        Saves all queries from the query dict and one more query (if passed);
        returns absolute path to passed query in file (if passed) or path of resulted file

        Parameters
        ----------
        query: otq.GraphQuery or None
            Additional query that is saved with all the queries from the storage. Usually, this will be the main query
            of the Source object that uses this storage
        query_name: str
            Name with which additional query will be saved. As additional query is not stored in any TmpOtq object,
            it can be saved with a name that is specified completely and not generated to be unique
        file_path: str or None
            Path to the file where all queries will be saved. If None, a TmpFile will be created
        file_suffix: str
            Only used if file_path is None. A suffix that is added to the name of a generated file.
        start: :py:class:`otp.datetime <onetick.py.datetime>`
            start time for the resulting .otq file
        end: :py:class:`otp.datetime <onetick.py.datetime>`
            end time for the resulting .otq file
        start_time_expression: str or None
            start time expression for the resulting .otq file
        end_time_expression: str or None
            end time expression for the resulting .otq file
        timezone: str
            timezone for the resulting .otq file
        symbol_date: :py:class:`otp.datetime <onetick.py.datetime>` or :py:class:`datetime.datetime` or int
            Symbol date for the query or integer in the YYYYMMDD format.
            Will be applied only to the query specified in the ``query`` parameter.

        Returns
        -------
            result: str , contains path to the resulting file with the name of the additional query
            (if "query" parameter is not None) or path to the resulting file otherwise

        """
        if file_path is None:
            base_dir = None
            if os.getenv('OTP_WEBAPI_TEST_MODE'):
                base_dir = _tmp_otq_path()
            kwargs = {}
            if config.otq_debug_mode:
                kwargs['clean_up'] = False
            tmp_file = utils.TmpFile(suffix=file_suffix,
                                     base_dir=base_dir,
                                     **kwargs)
            file_path = tmp_file.path

        queries_dict = dict(self.queries)

        if query is not None:
            if query_name in self.queries.keys():
                query_name = self.name_generator.get_str() + "_" + query_name
            query_params = {}
            if running_query_flag:
                query_params['running_query_flag'] = running_query_flag
            if symbol_date is not None:
                query_params['symbol_date'] = symbol_date
            queries_dict[query_name] = (query, query_params)

        # defining file-wise start/end times and time expressions

        # timezone definitions
        # in onetick and in dateutil can differ (e.g. "GMT-10" is +10:00 offset in onetick and -10:00 offset
        # in dateutil), therefore we always use time expressions to force onetick to make time conversions by itself
        # TODO: create a BDS ticket for the onetick team to fix it somehow
        if is_datetime_type(start):
            if not start_time_expression:
                start_time_expression = datetime2expr(start)
            start = None
        if is_datetime_type(end):
            if not end_time_expression:
                end_time_expression = datetime2expr(end)
            end = None

        # constructing a list of otq.Query objects, which will hold graphs, names, start/end times etc.
        query_list = []
        for stored_query_name in queries_dict.keys():   # noqa
            stored_query = otq.Query(queries_dict[stored_query_name][0])
            stored_query_params = queries_dict[stored_query_name][1]
            if timezone:
                stored_query.set_timezone(timezone)
            stored_query.set_symbols(queries_dict[stored_query_name][0].symbols())
            stored_query.set_query_name(stored_query_name)
            stored_query.set_start_time(start)
            stored_query.set_end_time(end)
            if 'running_query_flag' in stored_query_params.keys():
                stored_query.set_running_query_flag(stored_query_params['running_query_flag'])
            if start_time_expression:
                stored_query.set_start_time_expression(start_time_expression)
            if end_time_expression:
                stored_query.set_end_time_expression(end_time_expression)
            if 'symbol_date' in stored_query_params.keys():
                stored_symbol_date = stored_query_params['symbol_date']
                if stored_symbol_date is not None:
                    stored_symbol_date = int(utils.symbol_date_to_str(stored_symbol_date))
                    stored_query.set_symbol_date(stored_symbol_date)
            query_list.append(stored_query)

        _ = otli.OneTickLib()
        otq_file = otq.OtqFile(query_list)
        otq_file.save_to_file(file_path)
        get_debug_logger().debug(f'otq file saved to: {file_path}')

        if query is not None:
            return file_path + "::" + query_name
        else:
            return file_path



import warnings
import datetime
from typing import TYPE_CHECKING, Optional, Set, Type, Union
from onetick.py.backports import Literal

from onetick import py as otp
from onetick.py import configuration
from onetick.py.core.column import _Column, field_name_contains_lowercase
from onetick.py.otq import otq
from onetick.py.utils import adaptive
from onetick.py.compatibility import is_save_snapshot_database_parameter_supported

from .misc import inplace_operation

if TYPE_CHECKING:
    from onetick.py.core.source import Source


@inplace_operation
def write(
    self,
    db: Union[str, 'otp.DB'],
    symbol: Union[str, 'otp.Column', None] = None,
    tick_type: Union[str, 'otp.Column', None] = None,
    date: Union[datetime.date, Type[adaptive], None] = adaptive,
    start: Optional[datetime.date] = None,
    end: Optional[datetime.date] = None,
    append: bool = False,
    keep_symbol_and_tick_type: Union[bool, Type[adaptive]] = adaptive,
    propagate: bool = True,
    out_of_range_tick_action: Literal['exception', 'ignore', 'load'] = 'exception',
    timestamp: Optional['otp.Column'] = None,
    keep_timestamp: bool = True,
    correction_type: Optional['otp.Column'] = None,
    replace_existing_time_series: bool = False,
    allow_concurrent_write: bool = False,
    context: Union[str, Type[adaptive]] = adaptive,
    use_context_of_query: bool = False,
    inplace: bool = False,
    **kwargs,
) -> Optional['Source']:
    """
    Saves data result to OneTick database.

    Note
    ----
    This method does not save anything. It adds instruction in query to save.
    Data will be saved when query will be executed.

    Using ``start``+``end`` parameters instead of single ``date`` have some limitations:

        * ``inplace`` is not supported
        * if ``DAY_BOUNDARY_TZ`` and ``DAY_BOUNDARY_OFFSET`` specified against
            individual locations of database, then day boundary could be calculated incorrectly.
        * ``out_of_range_tick_action`` could be only ``exception`` or ``ignore``

    Parameters
    ----------
    db: str or :py:class:`otp.DB <onetick.py.DB>`
        database name or object.
    symbol: str or Column
        resulting symbol name string or column to get symbol name from.
        If this parameter is not set, then ticks _SYMBOL_NAME pseudo-field is used.
        If it is empty, an attempt is made to retrieve
        the symbol name from the field named SYMBOL_NAME.
    tick_type: str or Column
        resulting tick type string or column to get tick type from.
        If this parameter is not set, the _TICK_TYPE pseudo-field is used.
        If it is empty, an attempt is made to retrieve
        the tick type from the field named TICK_TYPE.
    date: :py:class:`otp.datetime <onetick.py.datetime>` or None
        date where to save data.
        Should be set to `None` if writing to accelerator or memory database.
        By default, it is set to `otp.config.default_date`.
    start: :py:class:`otp.datetime <onetick.py.datetime>` or None
        Start date for data to save. It is inclusive.
        Cannot be used with ``date`` parameter.
        Also cannot be used with ``inplace`` set to ``True``.
        Should be set to `None` if writing to accelerator or memory database.
        By default, None.
    end: :py:class:`otp.datetime <onetick.py.datetime>` or None
        End date for data to save. It is exclusive, so be sure to set
        it to the next day after the last day in data.
        Cannot be used with ``date`` parameter.
        Also cannot be used with ``inplace`` set to ``True``.
        Should be set to `None` if writing to accelerator or memory database.
        By default, None.
    append: bool
        If False - data will be rewritten for this ``date``
        or range of dates (from ``start`` to ``end``),
        otherwise data will be appended: new symbols are added,
        existing symbols can be modified (append new ticks, modify existing ticks).
        This option is not valid for accelerator databases.
    keep_symbol_and_tick_type: bool
        keep fields containing symbol name and tick type when writing ticks
        to the database or propagating them.
        By default, this parameter is adaptive.
        If ``symbol`` or ``tick_type`` are column objects, then it's set to True.
        Otherwise, it's set to False.
    propagate: bool
        Propagate ticks after that event processor or not.
    out_of_range_tick_action: str
        Action to be executed if tick's timestamp's date is not ``date`` or between ``start`` or ``end``:

            * `exception`: runtime exception will be raised
            * `ignore`: tick will not be written to the database
            * `load`: writes tick to the database anyway.
                Can be used only with ``date``, not with ``start``+``end``.

        Default: `exception`
    timestamp: Column
        Field that contains the timestamp with which the ticks will be written to the database.
        By default, the TIMESTAMP pseudo-column is used.
    keep_timestamp: bool
        If ``timestamp`` parameter is set and this parameter is set to True,
        then timestamp column is removed.
    correction_type: Column
        The name of the column that contains the correction type.
        This column will be removed.
        If this parameter is not set, no corrections will be submitted.
    replace_existing_time_series: bool
        If ``append`` is set to True, setting this option to True instructs the loader
        to replace existing time series, instead of appending to them.
        Other time series will remain unchanged.
    allow_concurrent_write: bool
        Allows different queries running on the same server to load concurrently into the same database.
    context: str
        The server context used to look up the database.
        By default, `otp.config.context` is used if ``use_context_of_query`` is not set.
    use_context_of_query: bool
        If this parameter is set to True and the ``context`` parameter is not set,
        the context of the query is used instead of the default value of the ``context`` parameter.
    inplace: bool
        A flag controls whether operation should be applied inplace.
        If ``inplace=True``, then it returns nothing.
        Otherwise, method returns a new modified object.
        Cannot be ``True`` if ``start`` and ``end`` are set.
    kwargs:
        .. deprecated:: 1.21.0

        Use named parameters instead.

    Returns
    -------
    :class:`Source` or None

    See also
    --------
    **WRITE_TO_ONETICK_DB** OneTick event processor

    Examples
    --------
    >>> data = otp.Ticks(X=[1, 2, 3])
    >>> data = data.write('SOME_DB', symbol='S_WRITE', tick_type='T_WRITE')
    >>> otp.run(data)
                         Time  X
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.001  2
    2 2003-12-01 00:00:00.002  3
    >>> data = otp.DataSource('SOME_DB', symbol='S_WRITE', tick_type='T_WRITE')
    >>> otp.run(data)
                         Time  X
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.001  2
    2 2003-12-01 00:00:00.002  3
    """
    if 'append_mode' in kwargs:
        warnings.warn("Parameter 'append_mode' is deprecated, use 'append'", FutureWarning)
        append = kwargs.pop('append_mode')

    if 'timestamp_field' in kwargs:
        warnings.warn("Parameter 'timestamp_field' is deprecated, use 'timestamp'", FutureWarning)
        timestamp = kwargs.pop('timestamp_field')

    if 'keep_timestamp_field' in kwargs:
        warnings.warn("Parameter 'keep_timestamp_field' is deprecated, use 'keep_timestamp'", FutureWarning)
        keep_timestamp = kwargs.pop('keep_timestamp_field')

    if kwargs:
        raise TypeError(f'write() got unexpected arguments: {list(kwargs)}')

    kwargs = {}

    # validate field names
    for field_name in self.schema:
        if field_name_contains_lowercase(field_name):
            if otp.config.allow_lowercase_in_saved_fields:
                warnings.warn(
                    f'Field "{field_name}" contains lowercase characters and is being saved'
                    ' to a Onetick database. This field will be converted to uppercase upon saving.'
                )
            else:
                raise ValueError(
                    f'Field "{field_name}" contains lowercase characters and cannot be saved to a Onetick database'
                )

    if date is not adaptive and (start or end):
        raise ValueError('date cannot be used with start+end')

    if date is adaptive and (start and end) and inplace:
        # join_with_query and merge are used for multiple dates, so inplace is not supported
        raise ValueError('cannot run on multiple dates if inplace is True, use one value for date instead of start+end')

    if (start and not end) or (not start and end):
        raise ValueError('start and end should be both specified or both None')

    if date is adaptive:
        date = configuration.config.default_date

    if symbol is not None:
        if isinstance(symbol, _Column):
            kwargs['symbol_name_field'] = str(symbol)
            if keep_symbol_and_tick_type is adaptive:
                keep_symbol_and_tick_type = True
        else:
            kwargs.setdefault('symbol_name_field', '_SYMBOL_NAME_FIELD_')
            self[kwargs['symbol_name_field']] = symbol

    if tick_type is not None:
        if isinstance(tick_type, _Column):
            kwargs['tick_type_field'] = str(tick_type)
            if keep_symbol_and_tick_type is adaptive:
                keep_symbol_and_tick_type = True
        else:
            kwargs.setdefault('tick_type_field', '_TICK_TYPE_FIELD_')
            self[kwargs['tick_type_field']] = tick_type

    if keep_symbol_and_tick_type is adaptive:
        keep_symbol_and_tick_type = False

    if timestamp is not None:
        kwargs['timestamp_field'] = str(timestamp)

    if correction_type is not None:
        kwargs['correction_type_field'] = str(correction_type)

    if context is not adaptive:
        kwargs['context'] = context
    elif not use_context_of_query:
        if otp.config.context is not None:
            kwargs['context'] = otp.config.context

    if out_of_range_tick_action.upper() == 'IGNORE':
        # let's ignore
        pass
    elif out_of_range_tick_action.upper() == 'LOAD':
        if start and end:
            raise ValueError('LOAD out_of_range_tick_action cannot be used with start+end, use date instead')
    elif out_of_range_tick_action.upper() == 'EXCEPTION':
        if start and end:
            # WRITE_TO_ONETICK_DB use DAY_BOUNDARY_TZ and DAY_BOUNDARY_OFFSET
            # to check tick timestamp is out of range or not
            # so we mimic it here with THROW event processor
            src = otp.Source(otq.DbShowConfig(str(db), 'DB_TIME_INTERVALS'))
            src.table(inplace=True, DAY_BOUNDARY_TZ=str, DAY_BOUNDARY_OFFSET=int)
            # DAY_BOUNDARY_OFFSET offset are in seconds
            src['DAY_BOUNDARY_OFFSET'] = src['DAY_BOUNDARY_OFFSET'] * 1000
            src.rename(
                {'DAY_BOUNDARY_TZ': '__DAY_BOUNDARY_TZ', 'DAY_BOUNDARY_OFFSET': '__DAY_BOUNDARY_OFFSET'}, inplace=True
            )
            self = self.join_with_query(src, symbol=f"{str(db)}::DUMMY", caching='per_symbol')
            start_formatted = start.strftime('%Y-%m-%d')
            end_formatted = end.strftime('%Y-%m-%d')
            convert_timestamp = self['TIMESTAMP'].dt.strftime('%Y%m%d%H%M%S.%J', timezone=self['__DAY_BOUNDARY_TZ'])
            start_op = otp.dt(start).to_operation(timezone=self['__DAY_BOUNDARY_TZ']) + self['__DAY_BOUNDARY_OFFSET']
            self.throw(
                where=(self['TIMESTAMP'] < start_op),
                message=(
                    'Timestamp '
                    + convert_timestamp
                    + ' of a tick, visible or hidden, '
                    + f'earlier than {start_formatted} in timezone '
                    + self['__DAY_BOUNDARY_TZ']
                ),
                inplace=True,
            )
            end_op = otp.dt(end).to_operation(timezone=self['__DAY_BOUNDARY_TZ']) + self['__DAY_BOUNDARY_OFFSET']
            self.throw(
                where=(self['TIMESTAMP'] >= end_op),
                message=(
                    'Timestamp '
                    + convert_timestamp
                    + ' of a tick, visible or hidden, '
                    + f'later than {end_formatted} in timezone '
                    + self['__DAY_BOUNDARY_TZ']
                ),
                inplace=True,
            )
    else:
        raise ValueError(
            f'Unknown out_of_range_tick_action: {out_of_range_tick_action}.'
            ' Possible values are: "ignore", "exception"'
        )

    kwargs = dict(
        **kwargs,
        database=str(db),
        append_mode=append,
        keep_symbol_name_and_tick_type=keep_symbol_and_tick_type,
        keep_timestamp_field=keep_timestamp,
        replace_existing_time_series=replace_existing_time_series,
        allow_concurrent_write=allow_concurrent_write,
        use_context_of_query=use_context_of_query,
    )

    if start and end:
        days = (end - start).days
        if days < 0:
            raise ValueError("Parameter 'start' must be less than parameter 'end'")
        if days == 0:
            raise ValueError("Parameters 'start' and 'end' must specify different dates")
        branches = []
        for i in range(days):
            branch = self.copy()
            branch.sink(
                otq.WriteToOnetickDb(
                    date=(start + otp.Day(i)).strftime('%Y%m%d'),
                    propagate_ticks=propagate,
                    out_of_range_tick_action='IGNORE',
                    **kwargs,
                )
            )
            branches.append(branch)
        self = otp.merge(branches)
    else:
        self.sink(
            otq.WriteToOnetickDb(
                date=date.strftime('%Y%m%d') if date else '',  # type: ignore[union-attr]
                propagate_ticks=propagate,
                out_of_range_tick_action=out_of_range_tick_action.upper(),
                **kwargs,
            )
        )

    for col in ('_SYMBOL_NAME_FIELD_', '_TICK_TYPE_FIELD_'):
        if col in self.schema:
            self.drop(col, inplace=True)

    to_drop: Set[str] = set()
    if not keep_symbol_and_tick_type:
        if 'symbol_name_field' in kwargs:
            to_drop.add(kwargs['symbol_name_field'])
        if 'tick_type_field' in kwargs:
            to_drop.add(kwargs['tick_type_field'])
    if not keep_timestamp and timestamp is not None and str(timestamp) not in {'Time', 'TIMESTAMP'}:
        to_drop.add(str(timestamp))
    if correction_type is not None:
        to_drop.add(str(correction_type))
    self.schema.set(**{k: v for k, v in self.schema.items() if k not in to_drop})
    return self


@inplace_operation
def write_parquet(
    self,
    output_path,
    compression_type="snappy",
    num_tick_per_row_group=1000,
    partitioning_keys="",
    propagate_input_ticks=False,
    inplace=False,
):
    """
    Writes the input tick series to parquet data file.

    Input must not have field 'time' as that field will also be added by the EP in the resulting file(s)

    Parameters
    ----------
    output_path: str
        Path for saving ticks to Parquet file.
        Partitioned: Path to the root directory of the parquet files.
        Non-partitioned: Path to the parquet file.
    compression_type: str
        Compression type for parquet files.
        Should be one of these: `gzip`, `lz4`, `none`, `snappy` (default), `zstd`.
    num_tick_per_row_group: int
        Number of rows per row group.
    partitioning_keys: list, str
        List of fields (`list` or comma-separated string) to be used as keys for partitioning.

        Setting this parameter will switch this EP to partitioned mode.

        In non-partitioned mode, if the path points to a file that already exists, it will be overridden.
        When partitioning is active:

        * The target directory must be empty
        * Key fields and their string values will be automatically URL-encoded to avoid conflicts with
            filesystem naming rules.

        Pseudo-fields '_SYMBOL_NAME' and '_TICK_TYPE' may be used as `partitioning_keys` and
        will be added to the schema automatically.
    propagate_input_ticks: bool
        Switches propagation of the ticks. If set to `True`, ticks will be propagated.
    inplace: bool
        A flag controls whether operation should be applied inplace.
        If ``inplace=True``, then it returns nothing. Otherwise method
        returns a new modified object.

    See also
    --------
    | **WRITE_TO_PARQUET** OneTick event processor
    | :py:class:`onetick.py.ReadParquet`

    Examples
    --------
    Simple usage:

    >>> data = otp.Ticks(A=[1, 2, 3])
    >>> data = data.write_parquet("/path/to/parquet/file")  # doctest: +SKIP
    >>> otp.run(data)  # doctest: +SKIP
    """
    if not hasattr(otq, "WriteToParquet"):
        raise RuntimeError("Current version of OneTick don't support WRITE_TO_PARQUET EP")

    if isinstance(partitioning_keys, list):
        partitioning_keys = ",".join(partitioning_keys)

    compression_type = compression_type.upper()

    ep_kwargs = {}
    if 'num_tick_per_row_group' in otq.WriteToParquet.Parameters.list_parameters():
        ep_kwargs['num_tick_per_row_group'] = num_tick_per_row_group
    else:
        ep_kwargs['num_ticks_per_row_group'] = num_tick_per_row_group

    self.sink(
        otq.WriteToParquet(
            output_path=output_path,
            compression_type=compression_type,
            partitioning_keys=partitioning_keys,
            propagate_input_ticks=propagate_input_ticks,
            **ep_kwargs,
        )
    )

    return self


@inplace_operation
def save_snapshot(
    self: 'Source',
    snapshot_name='VALUE',
    snapshot_storage='memory',
    default_db='CEP_SNAPSHOT',
    database='',
    symbol_name_field=None,
    expected_symbols_per_time_series=1000,
    num_ticks=1,
    reread_prevention_level=1,
    group_by=None,
    expected_groups_per_symbol=10,
    keep_snapshot_after_query=False,
    allow_concurrent_writers=False,
    remove_snapshot_upon_start=None,
    inplace=False,
):
    """
    Saves last (at most) `n` ticks of each group of ticks from the input time series in global storage or
    in a memory mapped file under a specified snapshot name.
    Tick descriptor should be the same for all ticks saved into the snapshot.
    These ticks can then be read via :py:class:`ReadSnapshot <onetick.py.ReadSnapshot>` by using the name
    of the snapshot and the same symbol name (``<db_name>::<symbol>``) that were used by this method.

    The event processor cannot be used by default. To enable it, access control should be configured,
    so user could have rights to use **SAVE_SNAPSHOT** EP.

    Parameters
    ----------
    snapshot_name: str
        The name of the snapshot, can be any string which doesn't contain slashes or backslashes.
        Two snapshots can have the same name if they are stored in memory mapped files for different databases. Also,
        they can have the same names if they are stored in the memories of different processes (different tick_servers).
        In all other cases the names should be unique.

        Default: `VALUE`
    snapshot_storage: str
        This parameter specifies the place of storage of the snapshot. Possible options are:

        * `memory` - the snapshot is stored in the dynamic (heap) memory of the process
          that ran (or is still running) the :py:meth:`onetick.py.Source.save_snapshot` for the snapshot.
        * `memory_mapped_file` - the snapshot is stored in a memory mapped file.
          For each symbol to get the location of the snapshot in the file system, ``save_snapshot`` looks at
          the **SAVE_SNAPSHOT_DIR** parameter value in the locator section for the database of the symbol.
          In a specified directory it creates a new directory with the name of the snapshot and keeps
          the memory mapped file and some other helper files there.

        Default: `memory`
    default_db: str
        The ticks with empty symbol names or symbol names with no database name as a prefix are saved as
        if they have symbol names equal to **DEFAULT_DB::SYMBOL_NAME** (where **SYMBOL_NAME** can be empty).
        These kinds of ticks, for example, can appear after merging time series. To save/read these ticks
        to/from storage a dummy database with the specified default name should be configured in the locator.

        Default: `CEP_SNAPSHOT`
    database: str, optional
        Specifies the output database for saving the snapshot.
    symbol_name_field: str, :py:class:`~onetick.py.Column`, optional
        If this parameter is specified, then each input time series is assumed to be a union of several time series and
        the value of the specified attribute of each tick determines to which time series the tick actually belongs.
        These values should be pure symbol names (for instance if the tick belongs to the time series **DEMO_L1::A**,
        then the value of the corresponding attribute should be **A**) and the database name will be taken from
        symbol of the merged time series.
    expected_symbols_per_time_series: int
        This parameter makes sense only when ``symbol_name_field`` is specified.
        It is the number of real symbols that are expected to occur per input time series.
        Bigger numbers may result in larger memory utilization by the query but will make the query faster.

        Default: `1000`
    num_ticks: int
        The number of ticks to be stored for each group per each symbol.

        Default: `1`
    reread_prevention_level: int
        For better performance we do not use synchronization mechanisms between the snapshot writer[s] and reader[s].
        That is why when the writer submits ticks for some symbol very quickly the reader may fail to read
        those ticks, and it will keep trying to reread them until it succeeds.
        The ``reread_prevention_level`` parameter addresses this problem.
        The higher the reread prevention level the higher the chance for the reader to read ticks successfully.
        But high prevention level also means high memory utilization, that is why it is recommended to keep
        the value of this parameter unchanged until you get an error about inability of the reader to read the snapshot
        due to fast writer.

        Default: `1`
    group_by: list of str, :py:class:`~onetick.py.Column`, optional
        When specified, the EP will keep the last **n** ticks of each group for each symbol;
        otherwise it will just keep the last **n** ticks of the input time series.
        The group is a list of input ticks with the same values in the specified fields.
    expected_groups_per_symbol: int
        The number of expected groups of ticks for each time series.
        The specified value is used only when ``group_by`` fields are specified,
        otherwise it is ignored, and we assume that the number of expected groups is 1.
        The number hints the EP to allocate memory for such number of tick groups each time
        a new group of ticks is going to be created and no free memory is left.

        Default: `10`
    keep_snapshot_after_query: bool
        If the snapshot is saved in process memory and this parameter is set, the saved snapshot continues to live
        after the query ends. If this parameter is not set, the snapshot is removed as soon as the query finishes and
        its name is released for saving new snapshots with the same name.
        This parameter is ignored if the snapshot is saved in the memory mapped file.

        Default: `False`
    allow_concurrent_writers: bool
        If this parameter is ``True`` multiple saver queries can write to the same snapshot contemporaneously.
        But different writers should write to different time series.
        Also, saver queries should run inside the same process (i.e., different tick servers or loaders with otq
        transformers cannot write to the same ``memory_mapped_file`` snapshot concurrently).

        Default: `False`
    remove_snapshot_upon_start: bool, optional
        If this parameter is ``True`` the snapshot will be removed at the beginning of the query the next time
        ``save_snapshot`` is called for the same snapshot. If the parameter is ``False`` the snapshot
        with the specified name will be appended to upon the next run of ``save_snapshot``.

        If you'll leave this parameter as ``None``, it will be equal to setting this parameter to ``NOT_SET`` in EP.
        ``NOT_SET`` option operates in the same way as ``True`` for ``memory`` snapshots or ``False``
        for ``memory_mapped_file`` snapshots.

        Default: None (``NOT_SET``)
    inplace: bool
        A flag controls whether operation should be applied inplace.
        If ``inplace=True``, then it returns nothing. Otherwise method
        returns a new modified object.

    See also
    --------
    | **SAVE_SNAPSHOT** OneTick event processor
    | :py:class:`onetick.py.ReadSnapshot`
    | :py:class:`onetick.py.ShowSnapshotList`
    | :py:class:`onetick.py.FindSnapshotSymbols`
    | :py:meth:`onetick.py.Source.join_with_snapshot`

    Examples
    --------
    Save ticks to a snapshot in a memory:

    >>> src = otp.Ticks(X=[1, 2, 3, 4, 5])
    >>> src = src.save_snapshot(snapshot_name='some_snapshot')  # doctest: +SKIP
    >>> otp.run(src)  # doctest: +SKIP

    If you want to use snapshot, stored in memory, after query, use parameter ``keep_snapshot_after_query``:

    >>> src = src.save_snapshot(snapshot_name='some_snapshot', keep_snapshot_after_query=True)  # doctest: +SKIP

    Snapshot will be associated with default database. You can set database via ``database`` parameter:

    >>> src = src.save_snapshot(
    ...     snapshot_name='some_snapshot', database='SOME_DATABASE', keep_snapshot_after_query=True
    ... )  # doctest: +SKIP
    >>> otp.run(src)  # doctest: +SKIP
    >>>
    >>> src = otp.ShowSnapshotList()  # doctest: +SKIP
    >>> otp.run(src)  # doctest: +SKIP
            Time  SNAPSHOT_NAME STORAGE_TYPE        DB_NAME
    0 2003-12-01  some_snapshot       MEMORY  SOME_DATABASE

    By default, only one last tick per group, if it set, or from all ticks per symbol is saved.
    You can change this number by setting ``num_ticks`` parameter:

    >>> src = src.save_snapshot(snapshot_name='some_snapshot', num_ticks=100)  # doctest: +SKIP

    Setting symbol name for every tick in snapshot from source field:

    >>> src = otp.Ticks(X=[1, 2, 3], SYMBOL_FIELD=['A', 'B', 'C'])
    >>> src = src.save_snapshot(
    ...     snapshot_name='some_snapshot', symbol_name_field='SYMBOL_FIELD', keep_snapshot_after_query=True,
    ... )  # doctest: +SKIP
    >>> otp.run(src)  # doctest: +SKIP
    >>>
    >>> src = otp.FindSnapshotSymbols(snapshot_name='some_snapshot')  # doctest: +SKIP
    >>> otp.run(src)  # doctest: +SKIP
            Time SYMBOL_NAME
    0 2003-12-01  DEMO_L1::A
    1 2003-12-01  DEMO_L1::B
    2 2003-12-01  DEMO_L1::C

    Group ticks by column ``X`` and keep last 2 ticks from each group:

    >>> src = otp.Ticks(X=[0, 0, 0, 1, 1, 1], Y=[1, 2, 3, 4, 5, 6])
    >>> src = src.save_snapshot(
    ...     snapshot_name='some_snapshot', group_by=[src['X']], num_ticks=2, keep_snapshot_after_query=True,
    ... )  # doctest: +SKIP
    >>> otp.run(src)  # doctest: +SKIP
    >>>
    >>> src = otp.ReadSnapshot(snapshot_name='some_snapshot')  # doctest: +SKIP
    >>> otp.run(src)  # doctest: +SKIP
            Time  X  Y               TICK_TIME
    0 2003-12-01  0  2 2003-12-01 00:00:00.001
    1 2003-12-01  0  3 2003-12-01 00:00:00.002
    2 2003-12-01  1  5 2003-12-01 00:00:00.004
    3 2003-12-01  1  6 2003-12-01 00:00:00.005

    """
    kwargs = {}

    if not hasattr(otq, "SaveSnapshot"):
        raise RuntimeError("Current version of OneTick doesn't support SAVE_SNAPSHOT EP")

    if snapshot_storage not in ['memory', 'memory_mapped_file']:
        raise ValueError('`snapshot_storage` must be one of "memory", "memory_mapped_file"')

    if isinstance(symbol_name_field, _Column):
        symbol_name_field = str(symbol_name_field)
    if symbol_name_field and symbol_name_field not in self.schema:
        raise ValueError(f'Field "{symbol_name_field}" passed as `symbol_name_field` parameter is not in schema.')

    is_database_param_supported = is_save_snapshot_database_parameter_supported()

    if database:
        if not is_database_param_supported:
            raise RuntimeError("Current version of OneTick doesn't support `database` parameter on SAVE_SNAPSHOT EP")

        kwargs['database'] = database

    if symbol_name_field is None:
        symbol_name_field = ''

    if group_by is None:
        group_by = []

    if not isinstance(group_by, list):
        raise ValueError('`group_by` must be a list')

    result_group_by = []

    for column in group_by:
        item = column
        if isinstance(column, _Column):
            item = str(column)

        if item not in self.schema:
            raise ValueError(f'Field "{item}" passed as `group_by` parameter is not in schema.')

        result_group_by.append(item)

    snapshot_storage = snapshot_storage.upper()

    if remove_snapshot_upon_start is None:
        remove_snapshot_upon_start = 'NOT_SET'

    # clear schema
    self.schema.set()

    self.sink(
        otq.SaveSnapshot(
            snapshot_name=snapshot_name,
            snapshot_storage=snapshot_storage,
            default_db=default_db,
            symbol_name_field=symbol_name_field,
            expected_symbols_per_time_series=expected_symbols_per_time_series,
            num_ticks=num_ticks,
            reread_prevention_level=reread_prevention_level,
            group_by=','.join(result_group_by),
            expected_groups_per_symbol=expected_groups_per_symbol,
            keep_snapshot_after_query=keep_snapshot_after_query,
            allow_concurrent_writers=allow_concurrent_writers,
            remove_snapshot_upon_start=remove_snapshot_upon_start,
            **kwargs,
        )
    )

    return self


@inplace_operation
def write_text(
    self: 'Source',
    *,
    propagate_ticks=True,
    output_headers=True,
    output_types_in_headers=False,
    order=None,
    prepend_symbol_name=True,
    prepended_symbol_name_size=0,
    prepend_timestamp=True,
    separator=',',
    formats_of_fields=None,
    double_format='%f',
    output_dir=None,
    output_file=None,
    error_file=None,
    warning_file=None,
    data_quality_file=None,
    treat_input_as_binary=False,
    flush=True,
    append=False,
    allow_concurrent_write=False,
    inplace=False,
):
    r"""
    Writes the input tick series to a text file or standard output.

    Parameters
    ----------
    propagate_ticks: bool
        If True (default) then ticks will be propagated after this method, otherwise this method won't return ticks.
    output_headers: bool
        Switches the output of the headers.
        If True (default), a tick descriptor line appears in the output before the very first tick for that query.
        If the structure of the output tick changes, another tick descriptor line appears before the first changed tick.
        The header line starts with **#**.
        The field names are ordered as mandated by the ``order`` parameter or,
        if it is empty, in the order of appearance in the tick descriptor.
        Fields that are not specified in the ``order`` parameter
        will appear after specified ones in the order of their appearance in the tick descriptor.
    output_types_in_headers: bool
        Switches the output of field types in the header lines.
        ``output_types_in_headers`` can be set only when ``output_headers`` is set too.
    order: list
        The field appearance order in the output.
        If all or some fields are not specified,
        those fields will be written in the order of their appearance in the tick descriptor.

        Field **SYMBOL_NAME** may be specified if parameter ``prepend_symbol_name`` is set.

        Field **TIMESTAMP** may be specified if parameter ``prepend_timestamp`` is set.
    prepend_symbol_name: bool
        If True (default), prepends symbol name before other fields as a new field named **SYMBOL_NAME** in the header
        (if ``output_headers`` is set).
    prepended_symbol_name_size: int
        When ``prepend_symbol_name`` is set, symbol will be adjusted to this size.
        If set to 0 (default), no adjustment will be done.
    prepend_timestamp: bool
        If set (default), tick timestamps, formatted as *YYYYMMDDhhmmss.qqqqqq* in the GMT time zone,
        will be prepended to the output lines.
        Header lines, if present, will have **TIMESTAMP** as the first field name.
        The default output format for tick timestamps can be specified in the ``formats_of_fields`` parameter.
    separator: str
        The delimiter string. This doesn't have to be a single character.
        Escape sequences are allowed for **\\t** (tab), **\\\\** (\\ character) and **\\xHH** (hex codes).
        By default "," (comma) will be used.
    formats_of_fields: dict
        The dictionary of field names and their formatting specifications.
        The formatting specification is the same as in the standard C
        `printf <https://pubs.opengroup.org/onlinepubs/009695399/functions/printf.html>`_ function.

        For float and decimal fields **%f** and **%.[<precision>]f** formats are only supported,
        first one being the default an outputting 6 decimal digits.

        Also if the field format starts with **%|**,
        it means that this is a timestamp field and should be in the format **%|tz|time_format_spec**,
        where the *tz* is the time zone name (if not specified GMT will be used),
        and *time_format_spec* is a custom time format specification,
        which is the same as the one used by the
        `strftime <https://pubs.opengroup.org/onlinepubs/009695399/functions/strftime.html>`_ function.

        In addition, you can also use **%q** , **%Q** , **%k** and **%J** placeholders,
        which will be replaced by 3 and 2 sign milliseconds, 6 sign microseconds and 9 sign nanoseconds, respectively.

        **%#**, **%-**, **%U**, **%N** placeholders will be replaced by Unix timestamp, Unix timestamp in milliseconds,
        microseconds and nanoseconds, respectively.

        **%+** and **%~** placeholders will be replaced by milliseconds and nanoseconds passed since midnight.
    double_format: str
        This format will be used for fields that are holding double values
        if they are not specified in ``formats_of_fields``.
    output_dir: str
        If specified, all output (output, warning, error, and data quality) files will be redirected to it.
        If this directory does not exist, it will get created.
        By default, the current directory is used.
    output_file: str
        The output file name for generated text data.
        If not set, the standard output will be used.
        It is also possible to add symbol name, database name, tick type,
        date of tick and query start time to the file name.
        For this special placeholders should be used, which will be replaced with the appropriate values:

            * **%SYMBOL%** - will be replaced with symbol name,
            * **%DBNAME%** - with database name,
            * **%TICKTYPE%** - with tick type,
            * **%DATE%** - with date of tick,
            * **%STARTTIME%** - with start time of the query.

        .. note::
            In case of using placeholders the output of the data may be split into different files.
            For example when querying several days of data and using **%DATE%** placeholder,
            the file will be created for every day of the interval.

        This format is also available for ``error_file``, ``warning_file`` and ``data_quality_file`` input parameters.
    error_file: str
        The file name where all error messages are directed.
        If not set the standard error will be used.
    warning_file: str
        The file name where all warning messages are directed.
        If not set the standard error will be used.
    data_quality_file: str
        The file name where all data quality messages are directed.
        If not set the standard error will be used.
    treat_input_as_binary: bool
        Opens output file in binary mode to not modify content of ticks when printing them to the file.
        Also in this mode method prints no new line to the file after every tick write.
    flush: bool
        If True (default) then the output will be flushed to disk after every tick.

        .. note::
            Notice that while this setting makes results of the query recorded into a file without delay,
            making them immediately available to applications that read this file,
            it may slow down the query significantly.

    append: bool
        If set to True, will try to append data to files (output, error, warning, data_quality), instead of overwriting.
    allow_concurrent_write: bool
        Allows different queries running on the same server to write concurrently to the same files
        (output, error, warning, data_quality).
    inplace: bool
        A flag controls whether operation should be applied inplace.
        If ``inplace=True``, then it returns nothing. Otherwise method
        returns a new modified object.

    See also
    --------
    | **WRITE_TEXT** OneTick event processor
    | :py:meth:`onetick.py.Source.dump`

    Examples
    --------

    By default the text is written to the standard output:

    >>> data = otp.Ticks(A=[1, 2, 3])
    >>> write = data.write_text()
    >>> _ = otp.run(write)  # doctest: +SKIP
    #SYMBOL_NAME,TIMESTAMP,A
    AAPL,20031201050000.000000,1
    AAPL,20031201050000.001000,2
    AAPL,20031201050000.002000,3

    Output file can also be specified:

    >>> write = data.write_text(output_file='result.csv')
    >>> _ = otp.run(write)  # doctest: +SKIP
    >>> with open('result.csv') as f:  # doctest: +SKIP
    ...     print(f.read())  # doctest: +SKIP
    #SYMBOL_NAME,TIMESTAMP,A
    AAPL,20031201050000.000000,1
    AAPL,20031201050000.001000,2
    AAPL,20031201050000.002000,3

    Symbol name, timestamp of the tick and can be removed from the output:

    >>> write = data.write_text(prepend_timestamp=False,
    ...                         prepend_symbol_name=False)
    >>> _ = otp.run(write)  # doctest: +SKIP
    #A
    1
    2
    3

    The header can also be removed from the output:

    >>> write = data.write_text(output_headers=False)
    >>> _ = otp.run(write)  # doctest: +SKIP
    AAPL,20031201050000.000000,1
    AAPL,20031201050000.001000,2
    AAPL,20031201050000.002000,3

    The order of fields and separator character can be specified:

    >>> write = data.write_text(order=['A', 'TIMESTAMP'],
    ...                         separator='\t',
    ...                         prepend_symbol_name=False)
    >>> _ = otp.run(write)  # doctest: +SKIP
    #A  TIMESTAMP
    1   20031201050000.000000
    2   20031201050000.001000
    3   20031201050000.002000

    The formatting can be specified for each field:

    >>> write = data.write_text(formats_of_fields={
    ...     'TIMESTAMP': '%|GMT|%Y-%m-%d %H:%M:%S.%q',
    ...     'A': '%3d'
    ... })
    >>> _ = otp.run(write)  # doctest: +SKIP
    #SYMBOL_NAME,TIMESTAMP,A
    AAPL,2003-12-01 05:00:00.000,  1
    AAPL,2003-12-01 05:00:00.001,  2
    AAPL,2003-12-01 05:00:00.002,  3
    """
    if output_types_in_headers and not output_headers:
        raise ValueError("Parameter 'output_types_in_headers' can only be set together with 'output_headers'")

    order = order or []
    formats_of_fields = formats_of_fields or {}
    for field in list(order) + list(formats_of_fields):
        if prepend_symbol_name and field == 'SYMBOL_NAME':
            continue
        if not prepend_symbol_name and field == 'SYMBOL_NAME' and field not in self.schema:
            raise ValueError(
                "Field 'SYMBOL_NAME' can't be specified in 'order' parameter if 'prepend_symbol_name' is not set"
            )
        if not prepend_timestamp and field == 'TIMESTAMP':
            raise ValueError(
                "Field 'TIMESTAMP' can't be specified in 'order' parameter if 'prepend_timestamp' is not set"
            )
        if field not in self.schema:
            raise ValueError(f"Field '{field}' is not in schema")

    kwargs = dict(
        propagate_ticks=propagate_ticks,
        output_headers=output_headers,
        output_types_in_headers=output_types_in_headers,
        order='|'.join(order),
        prepend_symbol_name=prepend_symbol_name,
        prepended_symbol_name_size=prepended_symbol_name_size,
        prepend_timestamp=prepend_timestamp,
        # OneTick uses \ as an escape character,
        # so replacing a single \ character with two \\ characters to escape it in OneTick
        separator=separator.replace('\\', r'\\'),
        formats_of_fields='\n'.join(f'{k}={v}' for k, v in formats_of_fields.items()),
        double_format=double_format,
        output_dir=output_dir,
        output_file=output_file,
        error_file=error_file,
        warning_file=warning_file,
        data_quality_file=data_quality_file,
        treat_input_as_binary=treat_input_as_binary,
        flush=flush,
        append=append,
        allow_concurrent_write=allow_concurrent_write,
    )
    for k, v in kwargs.items():
        if v is None:
            # None values may not be supported by onetick.query
            kwargs[k] = ''

    self.sink(otq.WriteText(**kwargs))
    if not propagate_ticks:
        self.schema.set(**{})
    return self



from typing import List, Dict, Union, Optional, Callable, TYPE_CHECKING
from onetick.py.backports import Literal

from functools import wraps
from inspect import Signature, Parameter
from onetick.py.core.column import Column
from onetick.py.core.column_operations.base import Operation, OnetickParameter
from onetick.py.core._source._symbol_param import _SymbolParamColumn
from onetick.py.types import OTPBaseTimeOffset

if TYPE_CHECKING:
    from onetick.py.core.source import Source  # hack for annotations

from onetick.py.docs.docstring_parser import Docstring
from onetick.py.docs.utils import param_doc
from onetick.py.aggregations.high_low import HighTick, LowTick


_column_doc = param_doc(
    name='column',
    str_annotation='str or Column or Operation',
    desc='''
    String with the name of the column to be aggregated or :py:class:`~onetick.py.Column` object.
    :py:class:`~onetick.py.Operation` object can also be used -- in this case
    the results of this operation for each tick are aggregated
    (see example in :ref:`common aggregation examples <aggregations_funcs>`).
    ''',
    annotation=Union[str, Column, Operation]
)

_running_doc = param_doc(
    name='running',
    desc='''
    See :ref:`Aggregation buckets guide <buckets_guide>` to see examples of how this parameter works.

    Specifies if the aggregation will be calculated as a sliding window.
    ``running`` and ``bucket_interval`` parameters determines when new buckets are created.

    * ``running`` = True

      aggregation will be calculated in a sliding window.

      * ``bucket_interval`` = N (N > 0)

        Window size will be N. Output tick will be generated when tick "enter" window (**arrival event**) and
        when "exit" window (**exit event**)

      * ``bucket_interval`` = 0

        Left boundary of window will be set to query start time. For each tick aggregation will be calculated in
        the interval [start_time; tick_t] from query start time to the tick's timestamp (inclusive).

    * ``running`` = False (default)

      buckets partition the [query start time, query end time) interval into non-overlapping intervals
      of size ``bucket_interval`` (with the last interval possibly of a smaller size).
      If ``bucket_interval`` is set to **0** a single bucket for the entire interval is created.

      Note that in non-running mode OneTick unconditionally divides the whole time interval
      into specified number of buckets.
      It means that you will always get this specified number of ticks in the result,
      even if you have less ticks in the input data.

    Default: False
    ''',
    default=False,
    annotation=bool
)

_all_fields_doc = param_doc(
    name='all_fields',
    desc="""
    See :ref:`Aggregation buckets guide <buckets_guide>` to see examples of how this parameter works.

    * ``all_fields`` = True

      output ticks include all fields from the input ticks

      * ``running`` = True

      an output tick is created only when a tick enters the sliding window

      * ``running`` = False

      fields of first tick in bucket will be used

    * ``all_fields`` = False and ``running`` = True

      output ticks are created when a tick enters or leaves the sliding window.

    * ``all_fields`` = "when_ticks_exit_window" and ``running`` = True

      output ticks are generated only for exit events, but all attributes from the exiting tick are copied over
      to the output tick and the aggregation is added as another attribute.
    """,
    default=False,
    annotation=Union[bool, str]
)
_all_fields_with_policy_doc = param_doc(
    name='all_fields',
    desc="""
    See :ref:`Aggregation buckets guide <buckets_guide>` to see examples of how this parameter works.

    - If ``all_fields`` False - output tick will have only aggregation fields.

    - If ``all_fields`` is True and ``running`` is False - additional fields will be copied from the first
      tick in the bucket to the output tick.

    - If ``all_fields`` False and ``running`` True - output ticks are created when a tick enters or leaves the
      sliding window.

    - If ``all_fields`` True and ``running`` True - an output tick is generated only for arrival events,
      but all attributes from the input tick causing an arrival event are copied over to the output tick
      and the aggregation is added as another attribute.

    - If ``all_fields`` True and ``running`` "when_ticks_exit_window" -
      an output tick is generated only for exit events,
      but all attributes from the exiting tick are copied over to the output tick
      and the aggregation is added as another attribute.

    - If ``all_fields`` set to "first", "last", "high", or "low" - explicitly set tick selection policy for all
      fields values. For "high" and "low" "PRICE" field will be selected as an input.
      Otherwise, you will get the runtime error.
      If ``all_fields`` is set to one of these values, ``running`` can't be `True`.

    - If ``all_fields`` is aggregation ``HighTick`` or ``LowTick`` - set tick selection policy for all fields values to
      "high" or "low" accordingly. But instead of "PRICE" the field selected as input will be set as aggregation's
      first parameter.
    """,
    default=False,
    str_annotation=("True, False, 'when_ticks_exit_window', 'first', 'last', 'high', 'low', "
                    ":py:func:`~onetick.py.agg.high_tick`, :py:func:`~onetick.py.agg.low_tick`"),
)
_bucket_interval_doc = param_doc(
    name='bucket_interval',
    desc="""
    Determines the length of each bucket (units depends on ``bucket_units``).

    If :py:class:`~onetick.py.Operation` of bool type is passed, acts as ``bucket_end_condition``.

    Bucket interval can also be set as a *float* value
    if ``bucket_units`` is set to *seconds*.
    Note that values less than 0.001 (1 millisecond) are not supported.

    Bucket interval can be set via some of the :ref:`datetime offset objects <datetime_offsets>`:
    :py:func:`otp.Milli <onetick.py.Milli>`, :py:func:`otp.Second <onetick.py.Second>`,
    :py:func:`otp.Minute <onetick.py.Minute>`, :py:func:`otp.Hour <onetick.py.Hour>`,
    :py:func:`otp.Day <onetick.py.Day>`, :py:func:`otp.Month <onetick.py.Month>`.
    In this case you could omit setting ``bucket_units`` parameter.

    Bucket interval can also be set with integer :py:class:`~onetick.py.core.column_operations.base.OnetickParameter`
    or :py:meth:`symbol parameter <onetick.py.core._source.symbol.SymbolType.__getitem__>`.
    """,
    default=0,
    annotation=Union[int, float, Operation, OnetickParameter, _SymbolParamColumn, OTPBaseTimeOffset],
    str_annotation=('int or float or :py:class:`~onetick.py.Operation`'
                    ' or :py:class:`~onetick.py.core.column_operations.base.OnetickParameter`'
                    ' or :py:meth:`symbol parameter <onetick.py.core._source.symbol.SymbolType.__getitem__>`'
                    ' or :ref:`datetime offset object <datetime_offsets>`'),
)
_bucket_time_doc = param_doc(
    name='bucket_time',
    desc="""
    Control output timestamp.

    * **start**

      the timestamp  assigned to the bucket is the start time of the bucket.

    * **end**

      the timestamp assigned to the bucket is the end time of the bucket.
    """,
    default="end",
    annotation=Literal["start", "end"]
)
_bucket_units_doc_kwargs = dict(
    name='bucket_units',
    desc="""
    Set bucket interval units.

    By default, if ``bucket_units`` and ``bucket_end_condition`` not specified, set to **seconds**.
    If ``bucket_end_condition`` specified, then ``bucket_units`` set to **flexible**.

    If set to **flexible** then ``bucket_end_condition`` must be set.

    Note that **seconds** bucket unit doesn't take into account daylight-saving time of the timezone,
    so you may not get expected results when using, for example, 24 * 60 * 60 seconds as bucket interval.
    In such case use **days** bucket unit instead.
    See example in :py:func:`onetick.py.agg.sum`.
    """,
    default=None,
    annotation=Optional[Literal["seconds", "ticks", "days", "months", "flexible"]]
)
_bucket_units_doc = param_doc(**_bucket_units_doc_kwargs)
_bucket_units_ob_doc = param_doc(**{
    **_bucket_units_doc_kwargs,
    # OB eps do not support 'ticks' bucket units
    'annotation': Optional[Literal["seconds", "days", "months", "flexible"]]
})

_bucket_end_condition_doc = param_doc(
    name='bucket_end_condition',
    str_annotation='condition',
    desc='''
    An expression that is evaluated on every tick. If it evaluates to "True", then a new bucket is created.
    This parameter is only used if ``bucket_units`` is set to "flexible".

    Also can be set via ``bucket_interval`` parameter by passing :py:class:`~onetick.py.Operation` object.
    ''',
    annotation=Optional[Operation],
    default=None
)
_end_condition_per_group_doc = param_doc(
    name='end_condition_per_group',
    desc='''
    Controls application of ``bucket_end_condition`` in groups.

    * ``end_condition_per_group`` = True

      ``bucket_end_condition`` is applied only to the group defined by ``group_by``

    * ``end_condition_per_group`` = False

      ``bucket_end_condition`` applied across all groups

    This parameter is only used if ``bucket_units`` is set to "flexible".

    When set to True, applies to all bucketing conditions. Useful, for example, if you need to specify ``group_by``,
    and you want to group items first, and create buckets after that.
    ''',
    annotation=bool,
    default=False
)
_boundary_tick_bucket_doc = param_doc(
    name='boundary_tick_bucket',
    desc='''
    Controls boundary tick ownership.

    * **previous**

      A tick on which ``bucket_end_condition`` evaluates to "true" belongs to the bucket being closed.

    * **new**

      tick belongs to the new bucket.

    This parameter is only used if ``bucket_units`` is set to "flexible"
    ''',
    annotation=Literal["new", "previous"],
    default="new"
)
_group_by_doc = param_doc(
    name='group_by',
    str_annotation='list, str or expression',
    desc='''
    When specified, each bucket is broken further into additional sub-buckets based on specified field values.
    If :py:class:`~onetick.py.Operation` is used then GROUP_{i} column is added. Where i is index in group_by list.
    For example, if Operation is the only element in ``group_by`` list then GROUP_0 field will be added.
    ''',
    annotation=Optional[Union[List, str, Operation]],
    default=None
)
_groups_to_display_doc = param_doc(
    name='groups_to_display',
    desc='''
    Specifies for which sub-buckets (groups) ticks should be shown for each bucket interval.
    By default **all** groups are shown at the end of each bucket interval.
    If this parameter is set to **event_in_last_bucket**, only the groups that received at least one tick
    within a given bucket interval are shown.
    ''',
    annotation=Literal["all", "previous"],
    default="all",
)
_biased_doc = param_doc(
    name='biased',
    annotation=bool,
    desc='''
    Switches between biased and unbiased standard deviation calculation.
    ''',
    default=True
)
_n_doc = param_doc(
    name='n',
    default=1,
    desc='''
    Number of ticks to output''',
    annotation=int
)
_keep_timestamp_doc = param_doc(
    name='keep_timestamp',
    desc="""
    If True, timestamps of the output ticks are the same as timestamps of the original ticks.
    Otherwise, timestamps of the output ticks are determined by bucket_time, and original timestamps
    are put in the TICK_TIME field.
    """,
    annotation=bool,
    default=True
)
_time_series_type_common = dict(
    name='time_series_type',
    desc='''
    Controls initial value for each bucket

    * **event_ts**

      only ticks from current bucket used for calculations

    * **state_ts**

      * if there is a tick in bucket with timestamp = bucket start

        only ticks in bucket used for calculation max value
      * else

        latest tick from previous bucket included in current bucket
    ''',
    annotation=Literal["event_ts", "state_ts"],
)
_time_series_type_doc = param_doc(
    default="event_ts",
    **_time_series_type_common,
)
_time_series_type_w_doc = param_doc(
    default="state_ts",
    **_time_series_type_common,
)
_selection_doc = param_doc(
    name='selection',
    desc='''
    Controls the selection of the respective beginning or trailing part of ticks.
    ''',
    annotation=Literal["first", "last"],
    default="first"
)
_side_doc = param_doc(
    name='side',
    desc="""
    Specifies whether the function is to be applied to sell orders (ASK), buy orders (BID), or both (empty).
    """,
    annotation=Literal['ASK', 'BID'],
    default=None,
)
_max_levels_doc = param_doc(
    name='max_levels',
    desc="""
    Number of order book levels (between 1 and 100_000) that need to be computed.
    If empty, all levels will be computed.
    """,
    annotation=int,
    default=None,
)
_min_levels_doc = param_doc(
    name='min_levels',
    desc="""
    Minimum number of order book levels, if so many levels are available in the book,
    that need to be included into the computation when either ``max_depth_shares`` or/and ``max_depth_for_price``
    are specified
    """,
    annotation=int,
    default=None,
)
_max_depth_shares_doc = param_doc(
    name='max_depth_shares',
    desc="""
    The total number of shares (i.e., the combined SIZE across top several levels of the book)
    that determines the number of order book levels that need to be part of the order book computation.
    If that number of levels exceeds `max_levels`, only `max_levels` levels of the book will be computed.
    The shares in excess of `max_depth_shares`, from the last included level, are not taken into account.
    """,
    annotation=int,
    default=None,
)
_max_depth_for_price_doc = param_doc(
    name='max_depth_for_price',
    desc="""
    The multiplier, product of which with the price at the top level of the book determines maximum price distance
    from the top of the book for the levels that are to be included into the book.
    In other words, only bids at <top_price>*(1-`max_depth_for_price`) and above
    and only asks of <top_price>*(1+`max_depth_for_price`) and less will be returned.
    If the number of the levels that are to be included into the book, according to this criteria,
    exceeds `max_levels`, only `max_levels` levels of the book will be returned.
    """,
    annotation=float,
    default=None,
)
_min_levels_doc = param_doc(
    name='min_levels',
    desc="""
    Minimum number of order book levels, if so many levels are available in the book, that need to be included
    into the computation when either `max_depth_shares` or/and `max_depth_for_price` are specified.
    """,
    annotation=int,
    default=None,
)
_max_initialization_days_doc = param_doc(
    name='max_initialization_days',
    desc="""
    This parameter specifies how many days back book event processors should go in order to find
    the latest full state of the book.
    The query will not go back resulting number of days if it finds initial book state earlier.
    When book event processors are used after VIRTUAL_OB EP, this parameter should be set to 0.
    When set, this parameter takes precedence over the configuration parameter BOOKS.MAX_INITIALIZATION_DAYS.
    """,
    annotation=int,
    default=1,
)
_book_uncross_method_doc = param_doc(
    name='book_uncross_method',
    desc="""
    When set to "REMOVE_OLDER_CROSSED_LEVELS", all ask levels that have price lower or equal to
    the price of a new bid tick get removed from the book, and all bid levels that have price higher or equal
    to the price of a new ask tick get removed from the book.
    """,
    annotation=Literal['REMOVE_OLDER_CROSSED_LEVELS'],
    default=None,
)
_dq_events_that_clear_book_doc = param_doc(
    name='dq_events_that_clear_book',
    desc="""
    A list of names of data quality events arrival of which should clear the order book.
    """,
    annotation=List[str],
    default=None,
)
_best_ask_price_field_doc = param_doc(
    name='best_ask_price_field',
    desc="""
    If specified, this parameter represents the name of the field value of which represents the lowest ask price
    starting from which the book ask size is to be computed.
    This value would also be used as the top price, relative to which ``max_depth_for_price`` would be computed.
    """,
    annotation=Union[str, Column],
    default=None,
)
_best_bid_price_field_doc = param_doc(
    name='best_bid_price_field',
    desc="""
    If specified, this parameter represents the name of the field value of which represents the highest bid price
    starting from which the book bid size is to be computed.
    This value would also be used as the top price, relative to which ``max_depth_for_price`` would be computed.
    """,
    annotation=Union[str, Column],
    default=None,
)
_bucket_interval_ob_num_levels_doc = param_doc(
    name='bucket_interval',
    desc="""
    Determines the length of each bucket in seconds.

    Bucket interval can be set via :ref:`datetime offset objects <datetime_offsets>`
    like :py:func:`otp.Second <onetick.py.Second>`, :py:func:`otp.Minute <onetick.py.Minute>`,
    :py:func:`otp.Hour <onetick.py.Hour>`, :py:func:`otp.Day <onetick.py.Day>`.
    In this case it will be converted to seconds.
    """,
    default=0,
    annotation=Union[int, OTPBaseTimeOffset],
    str_annotation='int or :ref:`datetime offset object <datetime_offsets>`',
)
_identify_source_doc = param_doc(
    name='identify_source',
    desc="""
    When this parameter is set to "true" and the input stream is fed through the VIRTUAL_OB event processor
    (with the QUOTE_SOURCE_FIELDS parameter specified) and `group_by` is not set to be "SOURCE"
    it will separate a tick with the same price from different sources into multiple ticks.
    The parameter can also be used when merging ticks from multiple feeds.
    Each feed going into the merge would need an ADD_FIELD EP source value set for the VALUE parameter,
    where the value would be different for each leg.
    """,
    annotation=bool,
    default=False,
)
_show_full_detail_doc = param_doc(
    name='show_full_detail',
    desc="""
    When set to "true" and if the state key of the input ticks consists of some fields besides PRICE,
    output ticks will contain all fields from the input ticks for each price level.
    When set to "false" only PRICE, UPDATE_TIME, SIZE, LEVEL, and BUY_SELL_FLAG fields will be populated.
    Note: setting this flag to "true" has no effect on a time series that does not have a state key.
    """,
    annotation=bool,
    default=False,
)
_show_only_changes_doc = param_doc(
    name='show_only_changes',
    desc="""
    When set to true, the output stream carries only changes to the book. The representation is as follows:
      * Changed and added levels are represented by themselves.
      * Deleted levels are shown with a size and level of zero.

    As with other modes, correct detection of update boundaries may require setting the `book_delimiters` option.
    """,
    annotation=bool,
    default=False,
)
_book_delimiters_doc = param_doc(
    name='book_delimiters',
    desc="""
    When set to "D" an extra tick is created after each book.
    Also, an additional column, called DELIMITER, is added to output ticks.
    The extra tick has values of all fields set to the defaults (0,NaN,""),
    except the delimiter field, which is set to "D."
    All other ticks have the DELIMITER set to zero (0).
    """,
    annotation=Literal['D'],
    default=None,
)
_state_key_max_inactivity_sec_doc = param_doc(
    name='state_key_max_inactivity_sec',
    desc="""
    If set, specifies in how many seconds after it was added
    a given state key should be automatically removed from the book.
    """,
    annotation=int,
    default=None,
)
_size_max_fractional_digits_doc = param_doc(
    name='size_max_fractional_digits',
    desc="""
    Specifies maximum number of digits after dot in SIZE, if SIZE can be fractional.
    """,
    annotation=int,
    default=0,
)
_include_market_order_ticks_doc = param_doc(
    name='include_market_order_ticks',
    desc="""
    If set, market order ticks (they have price NaN) are included into the order book,
    and are at the order book's top level.

        Default is False.
    """,
    annotation=bool,
    default=None,
)
_query_fun_doc = param_doc(
    name='query_fun',
    desc="""
    Function that takes :class:`~onetick.py.Source` as a parameter,
    applies some aggregation logic to it
    and returns :class:`~onetick.py.Source` as a result.
    Note that currently only methods that support dynamic symbol change
    could be used in the provided function.
    For example, :meth:`~onetick.py.Source.rename` can't be used.
    If you try to use such methods here, you will get an error during runtime.
    """,
    annotation=Callable,
)
_bucket_delimiter_doc = param_doc(
    name='bucket_delimiter',
    desc="""
    When set to ``True`` an extra tick is created after each bucket.
    Also, an additional column, called DELIMITER, is added to output ticks.
    The extra tick has values of all fields set to the defaults (0,NaN,""),
    except the delimiter field, which is set to "D"
    All other ticks have the DELIMITER set to string zero "0".
    """,
    annotation=bool,
    default=False,
)
_large_ints_doc = param_doc(
    name='large_ints',
    desc="""
    This parameter should be set
    if the input field of this aggregation may contain integer values that consist of 15 digits or more.

    Such large integer values cannot be represented by the double type without precision errors
    and thus require special handling.

    If set to **True** , the input field is expected to be a 64-bit integer number.
    The output field will also have 64-bit integer type.
    When no tick belongs to a given time bucket, the output value is set to a minimum of 64-bit integer.

    When this parameter set to :py:class:`onetick.py.adaptive`, the aggregation behaves the same way
    as when this parameter is set to True when the input field is a 64-bit integer type,
    and the same way as when this parameter is set to **False** when the input field is not a 64-bit integer type.
    """,
    annotation=bool,
    str_annotation="bool or :py:class:`onetick.py.adaptive`",
    default=False,
)
_null_int_val_doc = param_doc(
    name='null_int_val',
    desc="""
    The value of this parameter is considered to be the equivalent of ``NaN``
    when ``large_ints`` is set to ``True`` or
    when ``large_ints`` is set to :py:class:`onetick.py.adaptive` and the input field is a 64-bit integer type.
    """,
    annotation=int,
    default=0,
)
_skip_tick_if_doc = param_doc(
    name='skip_tick_if',
    desc="""
    If value of the input field is equal to the value in this parameter,
    this tick is ignored in the aggregation computation.

    This parameter is currently only supported for numeric fields.
    """,
    annotation=Optional[int],
    default=None,
)
_default_tick_doc = param_doc(
    name='default_tick',
    desc="""
    Mapping of input stream field names to some values.
    When set, exactly one tick with specified default values will be created for each empty bucket.
    If default value is specified as ``None``, then fields are initialized to the corresponding "zero" values:
    0 for integer types, NaN for doubles, empty string for string types, etc.
    """,
    annotation=Optional[dict],
    default=None,
)
_decay_doc = param_doc(
    name='decay',
    desc="""
    Weight decay. If **decay_value_type** is set to ``lambda``,
    **decay** provides the value of the **Lambda** variable in the aforementioned formula.
    Otherwise, if **decay_value_type** is set to ``half_life_index``, **decay** specifies the necessary number
    of consecutive ticks, the first one of which would have twice less the weight of the last one.
    The **Lambda** value is then calculated using this number.
    """,
    annotation=float,
)
_decay_value_type_common = dict(
    name='decay_value_type',
    desc="""
    The decay value can specified either directly or indirectly, controlled respectively by
    **lambda** and **half_life_index** values of this parameter.
    """,
    annotation=Literal['lambda', 'half_life_index'],
)
_decay_value_type_doc = param_doc(
    default='lambda',
    **_decay_value_type_common,
)
_decay_value_type_hl_doc = param_doc(
    default='half_life_index',
    **_decay_value_type_common,
)
_degree_doc = param_doc(
    name='degree',
    annotation=int,
    desc='''
    The order (degree) of the standardized moment to compute, denoted as k in the description above.
    ''',
    default=3,
)
_weight_field_name_doc = param_doc(
    name='weight_field_name',
    str_annotation='Optional, str or Column',
    annotation=Optional[Union[str, Column]],
    desc='''
    The name of the field that contains the current value of weight for a member of the portfolio
    that contributed the tick.

    You can also specify weight through the value of symbol parameter ``WEIGHT``.

    If ``weight_field_name`` is specified, all ticks should have the field pointed by this parameter and the value
    of this field is used as the weight.

    If weights are not specified in any of these ways, and you are running a single-stage query,
    the weights take the default value ``1``.
    ''',
    default='',
)
_weight_multiplier_field_name_doc = param_doc(
    name='weight_multiplier_field_name',
    desc="""
    Name of the field, value from which is used for multiplying portfolio value result.
    """,
    annotation=str,
    default='',
)
_portfolio_side_doc = param_doc(
    name='side',
    desc="""
    When set to **long**, the price of the portfolio is computed only for the input time series with ``weight > 0``.

    When set to **short**, the price of the portfolio is computed only for the input time series with ``weight < 0``.

    When set to **both**, the price of the portfolio is computed for all input time series.
    """,
    annotation=Literal['long', 'short', 'both'],
    default='both',
)
_weight_type_doc = param_doc(
    name='weight_type',
    desc="""
    When set to ``absolute``, the portfolio price is computed as the sum of ``input_field_value*weight``
    across all members of the portfolio.

    When set to ``relative``, the portfolio price is computed as the sum of
    ``input_field_value*weight/sum_of_all_weights`` across all members of the portfolio.
    """,
    annotation=Literal['absolute', 'relative'],
    default='absolute',
)
_portfolios_query_doc = param_doc(
    name='portfolios_query',
    desc="""
    A mandatory parameter that the specifies server-side `.otq` file that is expected to return
    mandatory columns ``PORTFOLIO_NAME`` and ``SYMBOL_NAME``,
    as well as an optional columns ``WEIGHT``, ``FX_SYMBOL_NAME`` and ``FX_MULTIPLY``.

    For a local OneTick server you can pass `otp.Source` objects.
    """,
    str_annotation='str, :class:`Source`',
)
_portfolios_query_params_doc = param_doc(
    name='portfolios_query_params',
    desc="""
    An optional parameter that specifies parameters of the query specified in `portfolios_query`.
    """,
    annotation=Union[str, Dict[str, str]],
    default='',
)
_portfolio_value_field_name_doc = param_doc(
    name='portfolio_value_field_name',
    desc="""
    List of the names (string with comma-separated list or ``list`` of strings/``Columns``) of the output fields
    which contain computed values of the portfolio.

    The number of the field names must match the number of the field names listed in the `columns` parameter.
    """,
    annotation=Union[str, List[Union[str, Column]]],
    default='VALUE',
)
_columns_portfolio_doc = param_doc(
    name='columns',
    str_annotation='str or list of Column or str',
    desc='''
    A list of the names of the input fields for which portfolio value is computed.

    Could be set as a comma-separated list of the names or ``list`` of name strings/``Columns`` objects.
    ''',
    annotation=Union[str, List[Union[str, Column]]],
    default='PRICE',
)
_symbols_doc = param_doc(
    name='symbols',
    desc="""
    Symbol(s) from which data should be taken.
    """,
    str_annotation=('str, list of str, :class:`Source`, :class:`query`, :py:func:`eval query <onetick.py.eval>`, '
                    ':py:class:`onetick.query.GraphQuery`.'),
    default=None,
)
_interest_rate_doc = param_doc(
    name='interest_rate',
    desc='''
    The risk-free interest rate.

    Could be set via a tick field, by specifying name of that field as string or
    passing :py:class:`~onetick.py.Column` object.
    ''',
    annotation=Optional[Union[int, float, str, Column]],
    default=None,
)
_price_field_doc = param_doc(
    name='price_field',
    desc='''
    The name of the field carrying the price value.
    ''',
    annotation=Union[str, Column],
    default='PRICE',
)
_option_price_field_doc = param_doc(
    name='option_price_field',
    desc='''
    The name of the field carrying the option price value.
    ''',
    annotation=Union[str, Column],
    default='OPTION_PRICE',
)
_method_doc = param_doc(
    name='method',
    desc='''
    Allowed values are ``newton``, ``newton_with_fallback`` and ``bisections``.

    Choose between ``newton`` and ``bisections`` for finding successively better approximations
    to the implied volatility value.

    Choose ``newton_with_fallback`` to automatically fall back to ``bisections`` method
    when ``newton`` fails to converge.
    ''',
    annotation=str,
    default='newton',
)
_precision_doc = param_doc(
    name='precision',
    desc='''
    Precision of the implied volatility value.
    ''',
    annotation=float,
    default='1.0e-5',
)
_value_for_non_converge_doc = param_doc(
    name='value_for_non_converge',
    desc='''
    Allowed values are ``nan_val`` and ``closest_found_val``, where ``closest_found_val`` stands
    for the volatility value for which the difference between calculated option price and input option price is minimal.

    Choose between ``nan_val`` and ``closest_found_val`` as implied volatility value,
    when the root-finding method does not converge within the specified precision.
    ''',
    annotation=str,
    default='nan_val',
)
_option_type_field_doc = param_doc(
    name='option_type_field',
    desc='''
    Specifies name of the field, which carries the option type (either **CALL** or **PUT**).
    ''',
    annotation=Union[str, Column],
    default='',
)
_strike_price_field_doc = param_doc(
    name='strike_price_field',
    desc='''
    Specifies name of the field, which carries the strike price of the option.
    ''',
    annotation=Union[str, Column],
    default='',
)
_days_in_year_doc = param_doc(
    name='days_in_year',
    desc='''
    Specifies number of days in a year (say, 365 or 252 (business days, etc.).
    Used with ``days_till_expiration`` parameter to compute the fractional years till expiration.
    ''',
    annotation=int,
    default=365,
)
_days_till_expiration_field_doc = param_doc(
    name='days_till_expiration_field',
    desc='''
    Specifies name of the field, which carries number of days till expiration of the option.
    ''',
    annotation=Union[str, Column],
    default='',
)
_expiration_date_field_doc = param_doc(
    name='expiration_date_field',
    desc='''
    Specifies name of the field, which carries the expiration date of the option, in **YYYYMMDD** format.
    ''',
    annotation=Union[str, Column],
    default='',
)


class DocMetaclass(type):
    def __new__(mcs, name, bases, attrs, parameters: Optional[list] = None):
        cls = super().__new__(mcs, name, bases, attrs)
        doc = None
        if '__init__' in attrs and cls.__init__.__doc__:    # type: ignore
            doc = cls.__init__.__doc__  # type: ignore
        elif '__doc__' in attrs:
            doc = attrs['__doc__']
        if doc and parameters:
            doc = Docstring(doc)
            for param in parameters:
                doc['Parameters'] = param
            cls.__init__.__doc__ = doc.build()  # type: ignore

        return cls


def copy_method(obj, mimic=True, drop_examples=False):

    """
    Decorator to copy aggregation function as method

    We assume that this decorator will be used only with aggregations.

    Updates (same way as for dict) `obj` docstring with decorated function docstring
    Updates decorated function signature: [self] + `obj` signature
    if mimic is True - won't execute decorated function, will execute `obj` and apply self instead
    if mimic is False - will execute decorated function as is

    Parameters
    ----------
    obj:
        donor aggregation function
    mimic: bool, default=True
        flag to execute decorated function or not
    drop_examples: bool
        Can be used to drop examples from ``obj`` docstring.
    """

    doc = obj.__doc__ or ''
    params = [Parameter(name='self',
                        kind=Parameter.POSITIONAL_OR_KEYWORD)] + list(Signature.from_callable(obj).parameters.values())
    doc = Docstring(doc)
    if drop_examples and 'Examples' in doc.docstring:
        doc.docstring.pop('Examples')

    def _decorator(fun):
        @wraps(fun)
        def _inner(self, *args, **kwargs):
            if mimic:
                agg = obj(*args, **kwargs)
                return agg.apply(self)
            else:
                return fun(self, *args, **kwargs)
        fun_doc = fun.__doc__ or ''
        fun_doc = Docstring(fun_doc)
        doc.update(fun_doc)
        _inner.__signature__ = Signature(parameters=params, return_annotation='Source')
        _inner.__doc__ = doc.build()
        return _inner
    return _decorator


def copy_signature(obj, add_self=False, drop_parameters=None, return_annotation='Source'):
    """
    Decorator that copies signature of the callable ``obj`` to the decorated function.
    """
    drop_parameters = drop_parameters or []
    obj_parameters = list(
        param for param_name, param in Signature.from_callable(obj).parameters.items()
        if param_name not in drop_parameters
    )
    params = []
    if add_self:
        params.append(
            Parameter(name='self', kind=Parameter.POSITIONAL_ONLY)
        )
    params.extend(obj_parameters)

    def _decorator(fun):
        @wraps(fun)
        def _inner(*args, **kwargs):
            return fun(*args, **kwargs)
        _inner.__signature__ = Signature(parameters=params, return_annotation=return_annotation)
        return _inner
    return _decorator



import datetime as dt

from typing import Optional, Union, Type, List

from onetick.py.otq import otq
import pandas as pd

from onetick.py.core._source._symbol_param import _SymbolParamColumn
from onetick.py.core.source import Source
from onetick.py.core.column_operations.base import OnetickParameter

from .. import types as ott
from .. import utils, configuration

AdaptiveTickType = Union[str, OnetickParameter, _SymbolParamColumn, Type[utils.adaptive]]
AdaptiveDBType = Union[str, OnetickParameter, _SymbolParamColumn, Type[utils.adaptive], List[str]]


def get_start_end_by_date(date):
    if isinstance(date, (ott.datetime, ott.date)):
        start = date.start
        end = date.end
    elif isinstance(date, (dt.datetime, dt.date)):
        start = dt.datetime(date.year, date.month, date.day)
        end = start + dt.timedelta(days=1, milliseconds=-1)
    else:
        raise ValueError(f"Unsupported value of parameter 'date': {type(date)}")
    return start, end


def convert_tick_type_to_str(
    tick_type: Optional[AdaptiveTickType],
    db: Optional[AdaptiveDBType] = None,
) -> Optional[str]:
    if not isinstance(db, list) and not isinstance(db, (OnetickParameter, _SymbolParamColumn)):
        if tick_type is utils.adaptive:
            tick_type = 'ANY'

        if isinstance(tick_type, type) or tick_type is None:
            return None

    if db is utils.adaptive_to_default or db is utils.adaptive:
        # if default database is not set, tick type will be set without it
        # and symbols will have to be specified in otp.run
        db = configuration.config.get('default_db')

    str_db = str(db) if db is not None else ''
    str_tt = str(tick_type) if tick_type is not None else ''

    is_db_symbol_param = isinstance(db, _SymbolParamColumn)
    is_tt_symbol_param = isinstance(tick_type, _SymbolParamColumn)

    if isinstance(db, OnetickParameter):
        str_db = db.parameter_expression

    if isinstance(tick_type, OnetickParameter):
        str_tt = tick_type.parameter_expression

    if db is None:
        if is_tt_symbol_param:
            return f"expr({str_tt})"
        return str_tt

    if isinstance(db, list):
        str_db = "+".join(db)

    if tick_type is not None:
        if is_db_symbol_param:
            if is_tt_symbol_param:
                return f"expr({str_db} + '::' + {str_tt})"

            return f"expr({str_db} + '::{str_tt}')"

        if is_tt_symbol_param:
            return f"expr('{str_db}::' + {str_tt})"
        elif "::" not in str_db:
            return str_db + "::" + str_tt
    elif is_db_symbol_param:
        return f"expr({str_db})"

    return str_db


def update_node_tick_type(
    src: "Source",
    tick_type: Optional[AdaptiveTickType],
    db: Optional[AdaptiveDBType] = None,
):
    """
    Update tick type of the node of the source ``src`` according to ``db`` name and ``tick_type``.

    Adaptive tick type means that tick type value doesn't affect query results
    and, thus, we change it to some constant value.

    Parameters
    ----------
    src: Source
        source to set tick type on
    tick_type: Optional[AdaptiveTickType]
        string tick type or :py:class:`onetick.py.adaptive`
    db: Optional[AdaptiveDBType]
        optional db name or list of strings with db names and tick types
    """

    str_db = convert_tick_type_to_str(tick_type, db)
    if str_db is None:
        return

    src.tick_type(str_db)


def _common_passthrough_base_ep(db, tick_type):
    src = Source(otq.Passthrough(fields="SYMBOL_NAME,TICK_TYPE", drop_fields=True))
    update_node_tick_type(src, tick_type, db)
    return src


def default_date_converter(date):
    return pd.to_datetime(date, format='%Y%m%d%H%M%S.%f')


def to_timestamp_nanos(date, date_converter, tz):
    date = date_converter(date)
    if isinstance(date, ott.dt):
        date = date.ts
    else:
        date = pd.to_datetime(date)
    return date.tz_localize(tz)



import functools
from typing import TYPE_CHECKING, Any

from onetick import py as otp
from onetick.py import types as ott
from onetick.py.core.column_operations._methods.op_types import are_numerics, are_time
from onetick.py.core.column_operations._methods.op_types import are_strings

if TYPE_CHECKING:
    import onetick.py as otp


def mean(self: 'otp.Source', *columns) -> 'otp.Operation':
    """
    Get average value of the specified columns.

    Do not confuse it with :py:func:`otp.agg.average <onetick.py.agg.average>`,
    this method gets average of the columns' values from single row, not doing
    aggregation of all rows.

    Parameters
    ----------
    columns: str, :class:`Column`
        Columns names or columns objects.
        All columns must have compatible types.

    Returns
    -------
    :class:`~onetick.py.Operation`

    Examples
    --------

    Integers and floating point values can be mixed:

    >>> t = otp.Tick(A=1, B=3.3)
    >>> t['AVG'] = t.mean('A', 'B')
    >>> otp.run(t)
            Time  A    B   AVG
    0 2003-12-01  1  3.3  2.15

    You can get the average for datetime values too:

    >>> t = otp.Tick(START_TIME=otp.dt(2022, 1, 1), END_TIME=otp.dt(2023, 1, 1))
    >>> t['MID_TIME'] = t.mean('START_TIME', 'END_TIME')
    >>> otp.run(t, timezone='GMT')
            Time  START_TIME    END_TIME             MID_TIME
    0 2003-12-01  2022-01-01  2023-01-01  2022-07-02 12:00:00

    Note that things may get confusing for datetimes when
    the timezone with daylight saving time is used:

    >>> t = otp.Tick(START_TIME=otp.dt(2022, 1, 1), END_TIME=otp.dt(2023, 1, 1))
    >>> t['MID_TIME'] = t.mean('START_TIME', 'END_TIME')
    >>> otp.run(t, timezone='EST5EDT')
            Time  START_TIME    END_TIME             MID_TIME
    0 2003-12-01  2022-01-01  2023-01-01  2022-07-02 13:00:00
    """
    if not columns:
        raise ValueError('No columns were specified')

    dtypes = set()
    for column_name in map(str, columns):
        if column_name not in self.schema:
            raise ValueError(f"There is no '{column_name}' column in the schema")
        dtypes.add(self.schema[column_name])

    if not are_numerics(*dtypes) and not are_time(*dtypes):
        raise ValueError(
            'Only int, float and datetime columns are supported. Numeric and datetime columns should not be mixed.'
        )

    op: Any = None
    for column_name in map(str, columns):
        column = self[column_name]
        dtype = self.schema[column_name]
        if dtype is otp.msectime and otp.nsectime in dtypes:
            column = column.astype(otp.nsectime)
        if are_time(dtype):
            column = column.astype(int)
        if op is None:
            op = column
        else:
            op += column

    op = op / len(columns)

    dtype = ott.get_type_by_objects(dtypes)
    if are_time(dtype):
        # can't convert float to datetime, converting to int first
        op = op.astype(int)
    op = op.astype(dtype)

    return op


def unite_columns(self: 'otp.Source', sep="", *, apply_str=False) -> 'otp.Operation':
    """
    Join values of all columns into one string

    The method unite all fields to one string, just like python ``join`` method. All fields should be strings,
    otherwise the error will be generated. To change this behavior, ``apply_str=True`` argument should be specified,
    in this case all fields will be converted to string type before joining.

    Parameters
    ----------
    sep: str
        Separator between values, empty string be dafault.
    apply_str: bool
        If set every column will be converted to string during operation. False be default.

    Returns
    -------
    result: column
        Column with str type

    Examples
    --------

    >>> # OTdirective: snippet-name: Arrange.join columns as strings;
    >>> data = otp.Ticks(X=[1, 2, 3], A=["A", "A", "A"], B=["A", "B", "C"])
    >>> data["S_ALL"] = data.unite_columns(sep=",", apply_str=True)
    >>> data["S"] = data[["A", "B"]].unite_columns()
    >>> otp.run(data)[["S", "S_ALL"]]
        S  S_ALL
    0  AA  1,A,A
    1  AB  2,A,B
    2  AC  3,A,C
    """
    if apply_str:
        cols = (self[col].apply(str) for col in self.schema)
    else:
        not_str = [name for name, t in self.schema.items() if not are_strings(t)]
        if not_str:
            raise ValueError(
                f"All joining columns should be strings, while {', '.join(not_str)} "
                f"are not. Specify `apply_str=True` for automatic type conversion"
            )
        else:
            cols = (self[col] for col in self.schema)
    return functools.reduce(lambda x, y: x + sep + y, cols)
